"""Power Supply Devices."""

import numpy as _np

from .. import util as _util

from ..namesys import SiriusPVName as _SiriusPVName
from ..search import PSSearch as _PSSearch
from ..pwrsupply.csdev import Const as _Const, \
    MAX_WFMSIZE_FBP as _MAX_WFMSIZE_FBP, \
    MAX_WFMSIZE_OTHERS as _MAX_WFMSIZE_OTHERS
from ..pwrsupply.psctrl.pscstatus import PSCStatus as _PSCStatus

from .device import Device as _Device
from .timing import Trigger as _Trigger


class _PSDev(_Device):
    """Base Power Supply Device."""

    PWRSTATE = _PSCStatus.PWRSTATE

    _default_timeout = 0.5  # [s]
    _properties_common = (
        'PwrState-Sel', 'PwrState-Sts',
    )
    _properties_linac = (
        'Current-SP', 'Current-RB', 'Current-Mon',
    )
    _properties_magps = (
        'Current-SP', 'Current-RB', 'Current-Mon', 'CurrentRef-Mon',
        'OpMode-Sel', 'OpMode-Sts',
        'WfmUpdateAuto-Sel', 'WfmUpdateAuto-Sts',
        'CycleType-Sel', 'CycleType-Sts',
        'CycleNrCycles-SP', 'CycleNrCycles-RB',
        'Wfm-SP', 'Wfm-RB', 'WfmRef-Mon', 'Wfm-Mon',
        'ScopeDuration-SP', 'ScopeDuration-RB',
        'ScopeFreq-SP', 'ScopeFreq-RB',
        'CycleFreq-SP', 'CycleFreq-RB',
        'CycleAmpl-SP', 'CycleAmpl-RB',
        'CycleOffset-SP', 'CycleOffset-RB',
        'CycleAuxParam-SP', 'CycleAuxParam-RB',
        'CycleEnbl-Mon',
    )
    _properties_fbp = _properties_magps + (
        'SOFBMode-Sel', 'SOFBMode-Sts'
        )
    _properties_fc = (
        'AlarmsAmp-Mon', 'OpMode-Sel', 'OpMode-Sts',
        'CurrLoopKp-RB', 'CurrLoopKp-SP', 'CurrLoopTi-RB', 'CurrLoopTi-SP',
        'CurrLoopMode-Sts', 'CurrLoopMode-Sel',
        'CurrGain-RB', 'CurrGain-SP', 'CurrOffset-RB', 'CurrOffset-SP',
        'Current-RB', 'Current-SP', 'Current-Mon', 'CurrentRef-Mon',
        'TestLimA-RB', 'TestLimA-SP', 'TestLimB-RB', 'TestLimB-SP',
        'TestWavePeriod-RB', 'TestWavePeriod-SP',
        'Voltage-RB', 'Voltage-SP', 'Voltage-Mon',
        'VoltGain-RB', 'VoltGain-SP', 'VoltOffset-RB', 'VoltOffset-SP',
        'InvRespMatRowX-SP', 'InvRespMatRowX-RB',
        'InvRespMatRowY-SP', 'InvRespMatRowY-RB',
        'FOFBAccGain-SP', 'FOFBAccGain-RB',
        'FOFBAccFreeze-Sel', 'FOFBAccFreeze-Sts',
        'FOFBAccClear-Cmd',
        'FOFBAccSatMax-SP', 'FOFBAccSatMax-RB',
        'FOFBAccSatMin-SP', 'FOFBAccSatMin-RB',
        'FOFBAcc-Mon',
        'FOFBAccDecimation-SP', 'FOFBAccDecimation-RB',
    )
    _properties_pulsed = (
        'Voltage-SP', 'Voltage-RB', 'Voltage-Mon',
        'Pulse-Sel', 'Pulse-Sts')
    _properties_pulsed_sept = (
        'Intlk1-Mon', 'Intlk2-Mon', 'Intlk3-Mon', 'Intlk4-Mon',
        'Intlk5-Mon', 'Intlk6-Mon', 'Intlk7-Mon',
    )
    _properties_pulsed_kckr = _properties_pulsed_sept + ('Intlk8-Mon', )
    _properties_pulsed_nlkckr = _properties_pulsed_kckr + (
        'CCoilHVoltage-SP', 'CCoilHVoltage-RB', 'CCoilHVoltage-Mon',
        'CCoilVVoltage-SP', 'CCoilVVoltage-RB', 'CCoilVVoltage-Mon',
    )

    def __init__(self, devname, auto_monitor_mon=False, props2init='all'):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in _PSSearch.get_psnames():
            raise NotImplementedError(devname)

        # power supply type and magnetic function
        (self._pstype, self._psmodel, self._magfunc,
         self._strength_propty, self._strength_units,
         self._is_linac, self._is_pulsed, self._is_fc, self._is_fbp,
         self._is_magps) = _PSDev.get_device_type(devname)

        # set attributes
        (self._strength_sp_propty,
         self._strength_rb_propty,
         self._strength_mon_propty,
         properties) = self._set_attributes_properties(devname)

        if props2init == 'all':
            props2init = properties
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)

        # private attribute with strength setpoint pv object
        self._strength_sp_pv = self.pv_object(self._strength_sp_propty)

    @property
    def pstype(self):
        """Return type of magnet(s) excited by power supply device."""
        return self._pstype

    @property
    def psmodel(self):
        """Return power supply model of the device."""
        return self._psmodel

    @property
    def magfunc(self):
        """Return function of magnet excited by power supply devices."""
        return self._magfunc

    @property
    def is_linac(self):
        """Return True if device is a Linac magnet power supply."""
        return self._is_linac

    @property
    def is_pulsed(self):
        """Return True if device is a pulsed magnet power supply."""
        return self._is_pulsed

    @property
    def is_fc(self):
        """Return True if device is a Sirius fast corrector power supply"""
        return self._is_fc

    @property
    def is_magps(self):
        """Return True if device is a Sirius magnet power supply."""
        return self._is_pulsed

    @property
    def strength_property(self):
        """Return Strength name."""
        return self._strength_propty

    @property
    def strength_units(self):
        """Return Strength units."""
        return self._strength_units

    @property
    def strength(self):
        """Return Strength RB."""
        return self[self._strength_rb_propty]

    @strength.setter
    def strength(self, value):
        """Set Strength SP."""
        self[self._strength_sp_propty] = value

    @property
    def strengthref_mon(self):
        """Return Strength Ref-Mon."""
        return self[self._strength_propty + 'Ref-Mon']

    @property
    def strength_mon(self):
        """Return Strength Mon."""
        return self[self._strength_mon_propty]

    @property
    def strength_upper_ctrl_limit(self):
        """Return Strength SP upper control limit."""
        return self._strength_sp_pv.upper_ctrl_limit

    @property
    def strength_lower_ctrl_limit(self):
        """Return Strength SP lower control limit."""
        return self._strength_sp_pv.lower_ctrl_limit

    @property
    def strength_upper_alarm_limit(self):
        """Return Strength SP upper alarm limit."""
        return self._strength_sp_pv.upper_alarm_limit

    @property
    def strength_lower_alarm_limit(self):
        """Return Strength SP lower alarm limit."""
        return self._strength_sp_pv.lower_alarm_limit

    @property
    def strength_upper_warning_limit(self):
        """Return Strength SP upper warning limit."""
        return self._strength_sp_pv.upper_warning_limit

    @property
    def strength_lower_warning_limit(self):
        """Return Strength SP lower warning limit."""
        return self._strength_sp_pv.lower_warning_limit

    @property
    def strength_upper_disp_limit(self):
        """Return Strength SP upper display limit."""
        return self._strength_sp_pv.upper_disp_limit

    @property
    def strength_lower_disp_limit(self):
        """Return Strength SP lower display limit."""
        return self._strength_sp_pv.lower_disp_limit

    @property
    def pwrstate(self):
        """."""
        return self['PwrState-Sts']

    @pwrstate.setter
    def pwrstate(self, value):
        """."""
        self['PwrState-Sel'] = value

    def set_strength(self, value, tol=0.2, timeout=10, wait_mon=False):
        """Set strength and wait until it gets there."""
        self.strength = value
        pv2wait = self._strength_mon_propty if wait_mon \
            else self._strength_rb_propty
        return self._wait_float(pv2wait, value, abs_tol=tol, timeout=timeout)

    def cmd_turn_on(self, timeout=_default_timeout):
        """."""
        self.pwrstate = self.PWRSTATE.On
        return self._wait('PwrState-Sts', self.PWRSTATE.On, timeout=timeout)

    def cmd_turn_off(self, timeout=_default_timeout):
        """."""
        self.pwrstate = self.PWRSTATE.Off
        return self._wait('PwrState-Sts', self.PWRSTATE.Off, timeout=timeout)

    @staticmethod
    def get_device_type(devname):
        """."""
        pstype = _PSSearch.conv_psname_2_pstype(devname)
        psmodel = _PSSearch.conv_psname_2_psmodel(devname)
        magfunc = _PSSearch.conv_psname_2_magfunc(devname)
        strength_propty = _util.get_strength_label(magfunc)
        strength_units = _util.get_strength_units(magfunc, pstype)
        is_linac = devname.sec.endswith('LI')
        is_pulsed = devname.dis == 'PU'
        is_fc = devname.dev == 'FCH' or devname.dev == 'FCV'
        is_fbp = psmodel == 'FBP'
        is_magps = not is_linac and not is_pulsed and not is_fc and not is_fbp
        return (pstype, psmodel, magfunc,
                strength_propty, strength_units,
                is_linac, is_pulsed, is_fc, is_fbp, is_magps)

    # --- private methods ---

    def _set_attributes_properties(self, devname):

        properties = _PSDev._properties_common
        if self._is_linac:
            properties += _PSDev._properties_linac
        elif self._is_pulsed:
            properties += _PSDev._properties_pulsed
            if self._psmodel == 'FP_KCKR':
                properties += _PSDev._properties_pulsed_kckr
            else:
                properties += _PSDev._properties_pulsed_sept
            if 'NLKckr' in devname:
                properties += _PSDev._properties_pulsed_nlkckr
        elif self._is_fc:
            properties += _PSDev._properties_fc
        elif self._is_fbp:
            properties += _PSDev._properties_fbp
        else:
            properties += _PSDev._properties_magps

        # strength properties
        strength_sp_propty = self._strength_propty + '-SP'
        strength_rb_propty = self._strength_propty + '-RB'
        strength_mon_propty = self._strength_propty + '-Mon'
        properties += (
            strength_sp_propty,
            strength_rb_propty,
            strength_mon_propty,
        )
        if not self._is_linac and not self._is_pulsed:
            strengthref_mon_propty = self._strength_propty + 'Ref-Mon'
            properties += (strengthref_mon_propty, )

        ret = (
            strength_sp_propty, strength_rb_propty, strength_mon_propty,
            properties)

        return ret


class PowerSupply(_PSDev):
    """Power Supply Device."""

    OPMODE_SEL = _PSCStatus.OPMODE
    OPMODE_STS = _PSCStatus.STATES
    CYCLETYPE = _PSCStatus.CYCLETYPE
    WFMUPDATEAUTO = _Const.DsblEnbl

    class DEVICES:
        """Devices names."""

    @property
    def current(self):
        """."""
        return self['Current-RB']

    @current.setter
    def current(self, value):
        self['Current-SP'] = value

    @property
    def current_mon(self):
        """."""
        return self['Current-Mon']

    @property
    def currentref_mon(self):
        """."""
        return self['CurrentRef-Mon']

    @property
    def opmode(self):
        """."""
        return self['OpMode-Sts']

    @opmode.setter
    def opmode(self, value):
        self._enum_setter(
            'OpMode-Sel', value, self.OPMODE_SEL)

    @property
    def opmode_str(self):
        """."""
        return self.OPMODE_STS._fields[self['OpMode-Sts']]

    def wait_cycle_to_finish(self, timeout=10):
        """."""
        return self._wait('CycleEnbl-Mon', 0, timeout)

    @property
    def cycle_enabled(self):
        """."""
        return bool(self['CycleEnbl-Mon'])

    @property
    def cycle_type(self):
        """."""
        return self['CycleType-Sts']

    @cycle_type.setter
    def cycle_type(self, value):
        self._enum_setter(
            'CycleType-Sel', value, self.CYCLETYPE)

    @property
    def cycle_type_str(self):
        """."""
        return self.CYCLETYPE._fields[self['CycleType-Sts']]

    @property
    def cycle_num_cycles(self):
        """Return the number of cycles of the cycling curve."""
        return self['CycleNrCycles-RB']

    @cycle_num_cycles.setter
    def cycle_num_cycles(self, value):
        """Return the number of cycles of the cycling curve."""
        self['CycleNrCycles-SP'] = value

    @property
    def cycle_freq(self):
        """Frequency of the cycling curve [Hz]."""
        return self['CycleFreq-RB']

    @cycle_freq.setter
    def cycle_freq(self, value):
        """Frequency of the cycling curve [Hz]."""
        self['CycleFreq-SP'] = value

    @property
    def cycle_period(self):
        """Period of the cycling curve [s]."""
        freq = self['CycleFreq-RB']
        return 1 / freq if freq != 0 else float('nan')

    @cycle_period.setter
    def cycle_period(self, value):
        """Period of the cycling curve [s]."""
        if value == 0:
            raise ValueError('Cannot set zero for cycle period.')
        self['CycleFreq-SP'] = 1 / value

    @property
    def cycle_ampl(self):
        """Return the current amplitude of the cycling curve [A]."""
        return self['CycleAmpl-RB']

    @cycle_ampl.setter
    def cycle_ampl(self, value):
        """Return the current amplitude of the cycling curve [A]."""
        self['CycleAmpl-SP'] = value

    @property
    def cycle_offset(self):
        """Return the current offset of the cycling curve [A]."""
        return self['CycleOffset-RB']

    @cycle_offset.setter
    def cycle_offset(self, value):
        """Return the current offset of the cycling curve [A]."""
        self['CycleOffset-SP'] = value

    @property
    def cycle_aux_param(self):
        """Meaning of each index is presented below.

        for Sine and Square:
         - AuxParams[0] --> initial phase [°]
         - AuxParams[1] --> final phase [°]
         - AuxParams[2] --> not used
         - AuxParams[3] --> not used

        for DampedSine and DampedSquaredSine:
         - AuxParams[0] --> initial phase [°]
         - AuxParams[1] --> final phase [°]
         - AuxParams[2] --> damping time [s]
         - AuxParams[3] --> not used

        for Trapezoidal:
         - AuxParams[0] --> rampup time [s]
         - AuxParams[1] --> plateau time [s]
         - AuxParams[2] --> rampdown time [s]
         - AuxParams[3] --> not used
        """
        value = self['CycleAuxParam-RB']
        if value is not None:
            return value.copy()
        return None

    @cycle_aux_param.setter
    def cycle_aux_param(self, value):
        self['CycleAuxParam-SP'] = value

    @property
    def cycle_duration(self):
        """Total duration of the cycling process [s]."""
        if self.cycle_freq != 0:
            return self.cycle_num_cycles / self.cycle_freq
        return float('nan')

    @property
    def cycle_rampup_time(self):
        """Rampup time for Trapezoidal signals [s]."""
        return self.cycle_aux_param[0]

    @cycle_rampup_time.setter
    def cycle_rampup_time(self, value):
        """Set Rampup time for Trapezoidal signals [s]."""
        var = self.cycle_aux_param
        var[0] = value
        self.cycle_aux_param = var

    @property
    def cycle_theta_begin(self):
        """Return initial phase for Sine or Damped(Squared)Sine [°]."""
        return self.cycle_aux_param[0]

    @cycle_theta_begin.setter
    def cycle_theta_begin(self, value):
        """Set Initial phase for Sine or Damped(Squared)Sine [°]."""
        var = self.cycle_aux_param
        var[0] = value
        self.cycle_aux_param = var

    @property
    def cycle_rampdown_time(self):
        """Rampdown time for Trapezoidal signals [s]."""
        return self.cycle_aux_param[2]

    @cycle_rampdown_time.setter
    def cycle_rampdown_time(self, value):
        """Set Rampdown time for Trapezoidal signals [s]."""
        var = self.cycle_aux_param
        var[2] = value
        self.cycle_aux_param = var

    @property
    def cycle_theta_end(self):
        """Return final phase for Sine or Damped(Squared)Sine [°]."""
        return self.cycle_aux_param[1]

    @cycle_theta_end.setter
    def cycle_theta_end(self, value):
        """Set Final phase for Sine or Damped(Squared)Sine [°]."""
        var = self.cycle_aux_param
        var[1] = value
        self.cycle_aux_param = var

    @property
    def cycle_plateau_time(self):
        """Plateau time for Trapezoidal signals [s]."""
        return self.cycle_aux_param[1]

    @cycle_plateau_time.setter
    def cycle_plateau_time(self, value):
        """Set plateau time for Trapezoidal signals [s]."""
        var = self.cycle_aux_param
        var[1] = value
        self.cycle_aux_param = var

    @property
    def cycle_decay_time(self):
        """Decay time constant for Damped(Squared)Sine signals [s]."""
        return self.cycle_aux_param[2]

    @cycle_decay_time.setter
    def cycle_decay_time(self, value):
        """Set Decay time constant for Damped(Squared)Sine [s]."""
        var = self.cycle_aux_param
        var[2] = value
        self.cycle_aux_param = var

    @property
    def wfm(self):
        """."""
        return self['Wfm-RB'].copy()

    @wfm.setter
    def wfm(self, value):
        """."""
        value = _np.array(value).ravel()
        max_size = _MAX_WFMSIZE_OTHERS
        if self.psmodel == 'FBP':
            max_size = _MAX_WFMSIZE_FBP
        self['Wfm-SP'] = value[:max_size]

    @property
    def wfm_mon(self):
        """."""
        return self['Wfm-Mon'].copy()

    @property
    def wfmref_mon(self):
        """."""
        return self['WfmRef-Mon'].copy()

    @property
    def wfm_update_auto(self):
        """Waveform auto update."""
        return self['WfmUpdateAuto-Sts']

    @property
    def wfm_update_auto_str(self):
        """Waveform auto update."""
        return self.WFMUPDATEAUTO._fields[self['WfmUpdateAuto-Sts']]

    @wfm_update_auto.setter
    def wfm_update_auto(self, value):
        """Set waveform auto update."""
        self._enum_setter(
            'WfmUpdateAuto-Sel', value, self.WFMUPDATEAUTO)

    @property
    def scope_freq(self):
        """Scope frequency [Hz]."""
        return self['ScopeFreq-RB']

    @scope_freq.setter
    def scope_freq(self, value):
        """Set scope frequency [Hz]."""
        self['ScopeFreq-SP'] = float(value)

    @property
    def scope_duration(self):
        """Scope duration [s]."""
        return self['ScopeDuration-RB']

    @scope_duration.setter
    def scope_duration(self, value):
        """Set scope duration [s]."""
        self['ScopeDuration-SP'] = float(value)

    def cmd_slowref(self, timeout=_PSDev._default_timeout):
        """."""
        self['OpMode-Sel'] = self.OPMODE_SEL.SlowRef
        return self._wait(
            'OpMode-Sts', self.OPMODE_STS.SlowRef, timeout=timeout)

    def cmd_slowrefsync(self, timeout=_PSDev._default_timeout):
        """."""
        self['OpMode-Sel'] = self.OPMODE_SEL.SlowRefSync
        return self._wait(
            'OpMode-Sts', self.OPMODE_STS.SlowRefSync, timeout=timeout)

    def cmd_cycle(self, timeout=_PSDev._default_timeout):
        """."""
        self['OpMode-Sel'] = self.OPMODE_SEL.Cycle
        return self._wait(
            'OpMode-Sts', self.OPMODE_STS.Cycle, timeout=timeout)


class PowerSupplyPU(_PSDev):
    """Pulsed Power Supply Device."""

    PULSTATE = _PSCStatus.PWRSTATE
    DEF_TIMEOUT = 10

    class DEVICES:
        """Devices names."""

        TB_INJ_SEPT = 'TB-04:PU-InjSept'
        BO_INJ_KCKR = 'BO-01D:PU-InjKckr'
        BO_EJE_KCKR = 'BO-48D:PU-EjeKckr'
        TS_EJE_SEPTF = 'TS-01:PU-EjeSeptF'
        TS_EJE_SEPTG = 'TS-01:PU-EjeSeptG'
        TS_INJ_SEPTG_1 = 'TS-04:PU-InjSeptG-1'
        TS_INJ_SEPTG_2 = 'TS-04:PU-InjSeptG-2'
        TS_INJ_SEPTF = 'TS-04:PU-InjSeptF'
        SI_INJ_DPKCKR = 'SI-01SA:PU-InjDpKckr'
        SI_INJ_NLKCKR = 'SI-01SA:PU-InjNLKckr'
        SI_PING_H = 'SI-01SA:PU-PingH'
        SI_PING_V = 'SI-19C4:PU-PingV'
        ALL = (
            TB_INJ_SEPT,
            BO_INJ_KCKR, BO_EJE_KCKR,
            TS_EJE_SEPTF, TS_EJE_SEPTG,
            TS_INJ_SEPTG_1, TS_INJ_SEPTG_2,
            TS_INJ_SEPTF,
            SI_INJ_DPKCKR, SI_INJ_NLKCKR,
            SI_PING_H, SI_PING_V,
            )

    _properties_timing = ('Delay-SP', 'Delay-RB', 'DelayRaw-SP', 'DelayRaw-RB')

    def __init__(self, devname, props2init='all'):
        """."""
        # check if device exists
        if devname not in PowerSupplyPU.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, props2init=props2init)

        # create timing device
        self._dev_timing = self._create_timing_device()

    @property
    def voltage(self):
        """."""
        return self['Voltage-RB']

    @voltage.setter
    def voltage(self, value):
        self['Voltage-SP'] = value

    @property
    def voltage_mon(self):
        """."""
        return self['Voltage-Mon']

    @property
    def ccoilh_voltage(self):
        """."""
        return self['CCoilHVoltage-RB']

    @ccoilh_voltage.setter
    def ccoilh_voltage(self, value):
        self['CCoilHVoltage-SP'] = value

    @property
    def ccoilh_voltage_mon(self):
        """."""
        return self['CCoilHVoltage-Mon']

    @property
    def ccoilv_voltage(self):
        """."""
        return self['CCoilVVoltage-RB']

    @ccoilv_voltage.setter
    def ccoilv_voltage(self, value):
        self['CCoilVVoltage-SP'] = value

    @property
    def ccoilv_voltage_mon(self):
        """."""
        return self['CCoilVVoltage-Mon']

    @property
    def delay(self):
        """."""
        return self._dev_timing['Delay-RB']

    @delay.setter
    def delay(self, value):
        """."""
        self._dev_timing['Delay-SP'] = value

    @property
    def delay_raw(self):
        """."""
        return self._dev_timing['DelayRaw-RB']

    @delay_raw.setter
    def delay_raw(self, value):
        """."""
        self._dev_timing['DelayRaw-SP'] = value

    @property
    def pulse(self):
        """."""
        return self['Pulse-Sts']

    @pulse.setter
    def pulse(self, value):
        """."""
        self['Pulse-Sel'] = value

    def cmd_turn_on_pulse(self, timeout=DEF_TIMEOUT):
        """."""
        self.pulse = self.PULSTATE.On
        return self._wait('Pulse-Sts', value=self.PULSTATE.On, timeout=timeout)

    def cmd_turn_off_pulse(self, timeout=DEF_TIMEOUT):
        """."""
        self.pulse = self.PULSTATE.Off
        return self._wait(
            'Pulse-Sts', value=self.PULSTATE.Off, timeout=timeout)

    @property
    def properties(self):
        """Return device properties."""
        return self.properties_in_use + self._dev_timing.properties_in_use

    @property
    def pvnames(self):
        """Return device PV names."""
        return set(list(super().pvnames) +
                   list(self._dev_timing.pvnames))

    @property
    def interlock_ok(self):
        """Return whether all interlocks are in Ok state."""
        intlks = [p for p in self.properties_in_use if 'Intlk' in p]
        is_ok = True
        for ilk in intlks:
            is_ok &= self[ilk] == 1
        return is_ok

    @property
    def connected(self):
        """Return PVs connection status."""
        if not super().connected:
            return False
        return self._dev_timing.connected

    @property
    def disconnected_pvnames(self):
        """Return list of disconnected device PVs."""
        return set(list(super().disconnected_pvnames) +
                   list(self._dev_timing.disconnected_pvnames))

    def update(self):
        """Update device properties."""
        super().update()
        self._dev_timing.update()

    def pv_object(self, propty):
        """Return PV object for a given device property."""
        if propty in self._pvs:
            return super().pv_object(propty)
        return self._dev_timing.pv_object(propty)

    def pv_attribute_values(self, attribute):
        """Return property-value dict of a given attribute for all PVs."""
        attributes = super().pv_attribute_values(attribute)
        attributes_ti = self._dev_timing.pv_attribute_values(attribute)
        attributes.update(attributes_ti)
        return attributes

    def __getitem__(self, propty):
        """Return value of property."""
        if propty in self._pvs:
            return super().__getitem__(propty)
        return self._dev_timing[propty]

    def __setitem__(self, propty, value):
        """Set value of property."""
        if propty in self._pvs:
            super().__setitem__(propty, value)
        else:
            self._dev_timing[propty] = value

    # --- private methods ---

    def _create_timing_device(self):
        """."""
        devname = self._devname.substitute(dis='TI')
        return _Trigger(devname, props2init=PowerSupplyPU._properties_timing)


class PowerSupplyFC(_PSDev):
    """Fast Correctors Power Supply Device."""

    OPMODE_SEL = _Const.OpModeFOFBSel
    OPMODE_STS = _Const.OpModeFOFBSts

    class DEVICES:
        """Devices names."""

    @property
    def opmode(self):
        """OpMode."""
        return self['OpMode-Sts']

    @opmode.setter
    def opmode(self, value):
        self._enum_setter('OpMode-Sel', value, self.OPMODE_SEL)

    @property
    def current(self):
        """Current setpoint."""
        return self['Current-RB']

    @current.setter
    def current(self, value):
        self['Current-SP'] = value

    @property
    def current_mon(self):
        """Implemented current."""
        return self['Current-Mon']

    @property
    def currentref_mon(self):
        """Current reference."""
        return self['CurrentRef-Mon']

    @property
    def curr_gain(self):
        """Current gain for A<->raw unit conversion."""
        return self['CurrGain-RB']

    @curr_gain.setter
    def curr_gain(self, value):
        self['CurrGain-SP'] = value

    @property
    def curr_offset(self):
        """Current offset for A<->raw unit conversion."""
        return self['CurrOffset-RB']

    @curr_offset.setter
    def curr_offset(self, value):
        self['CurrOffset-SP'] = value

    @property
    def voltage(self):
        """."""
        return self['Voltage-RB']

    @voltage.setter
    def voltage(self, value):
        """."""
        self['Voltage-SP'] = value

    @property
    def voltage_mon(self):
        """."""
        return self['Voltage-Mon']

    @property
    def volt_gain(self):
        """."""
        return self['VoltGain-RB']

    @volt_gain.setter
    def volt_gain(self, value):
        """."""
        self['VoltGain-SP'] = value

    @property
    def volt_offset(self):
        """."""
        return self['VoltOffset-RB']

    @volt_offset.setter
    def volt_offset(self, value):
        """."""
        self['VoltOffset-SP'] = value

    @property
    def currloop_kp(self):
        """Current control loop Kp parameter."""
        return self['CurrLoopKp-RB']

    @currloop_kp.setter
    def currloop_kp(self, value):
        self['CurrLoopKp-SP'] = value

    @property
    def currloop_ti(self):
        """Current control loop Ti parameter."""
        return self['CurrLoopTi-RB']

    @currloop_ti.setter
    def currloop_ti(self, value):
        self['CurrLoopTi-SP'] = value

    @property
    def currloop_mode(self):
        """Current control loop mode."""
        return self['CurrLoopMode-Sts']

    @currloop_mode.setter
    def currloop_mode(self, value):
        self['CurrLoopMode-Sel'] = value

    @property
    def alarms_amp(self):
        """."""
        return self['AlarmsAmp-Mon']

    def cmd_opmode_manual(self, timeout=_PSDev._default_timeout):
        """Set opmode to manual."""
        return self._set_opmode(mode=self.OPMODE_SEL.manual, timeout=timeout)

    def cmd_opmode_fofb(self, timeout=_PSDev._default_timeout):
        """Set opmode to fofb."""
        return self._set_opmode(mode=self.OPMODE_SEL.fofb, timeout=timeout)

    def _set_opmode(self, mode, timeout):
        self['OpMode-Sel'] = mode
        return self._wait('OpMode-Sts', mode, timeout=timeout)

    @property
    def invrespmat_row_x(self):
        """Horizontal correction coefficient value."""
        return self['InvRespMatRowX-RB']

    @invrespmat_row_x.setter
    def invrespmat_row_x(self, value):
        self['InvRespMatRowX-SP'] = _np.array(value, dtype=float)

    @property
    def invrespmat_row_y(self):
        """Vertical correction coefficient value."""
        return self['InvRespMatRowY-RB']

    @invrespmat_row_y.setter
    def invrespmat_row_y(self, value):
        self['InvRespMatRowY-SP'] = _np.array(value, dtype=float)

    @property
    def fofbacc_gain(self):
        """FOFB accumulator gain."""
        return self['FOFBAccGain-RB']

    @fofbacc_gain.setter
    def fofbacc_gain(self, value):
        self['FOFBAccGain-SP'] = value

    @property
    def fofbacc_freeze(self):
        """FOFB accumulator freeze state."""
        return self['FOFBAccFreeze-Sts']

    @fofbacc_freeze.setter
    def fofbacc_freeze(self, value):
        self['FOFBAccFreeze-Sel'] = value

    @property
    def fofbacc_satmax(self):
        """FOFB accumulator maximum saturation."""
        return self['FOFBAccSatMax-RB']

    @fofbacc_satmax.setter
    def fofbacc_satmax(self, value):
        self['FOFBAccSatMax-SP'] = value

    @property
    def fofbacc_satmin(self):
        """FOFB accumulator minimum saturation."""
        return self['FOFBAccSatMin-RB']

    @fofbacc_satmin.setter
    def fofbacc_satmin(self, value):
        self['FOFBAccSatMin-SP'] = value

    @property
    def fofbacc_mon(self):
        """FOFB accumulator."""
        return self['FOFBAcc-Mon']

    @property
    def fofbacc_decimation(self):
        """FOFB accumulator decimation."""
        return self['FOFBAccDecimation-RB']

    @fofbacc_decimation.setter
    def fofbacc_decimation(self, value):
        self['FOFBAccDecimation-SP'] = value

    def cmd_fofbacc_clear(self):
        """Command to clear FOFB accumulator."""
        self['FOFBAccClear-Cmd'] = 1
        return True


class PowerSupplyFBP(PowerSupply):
    """FBP Power Supply Device."""

    SOFBMODE_SEL = _Const.DsblEnbl
    SOFBMODE_STS = _Const.DsblEnbl

    @property
    def sofbmode(self):
        """SOFB mode status."""
        return self['SOFBMode-Sts']

    def cmd_sofbmode_enable(self, timeout=_PSDev._default_timeout):
        """Command to enable SOFBMode. Send command and wait."""
        return self._cmd_sofbmode(
            timeout, self.SOFBMODE_SEL.Enbl, self.SOFBMODE_STS.Enbl)

    def cmd_sofbmode_disable(self, timeout=_PSDev._default_timeout):
        """Command to disable SOFBMode. Send command and wait."""
        return self._cmd_sofbmode(
            timeout, self.SOFBMODE_SEL.Dsbl, self.SOFBMODE_STS.Dsbl)

    def _cmd_sofbmode(self, timeout, state_sel, state_sts):
        self['SOFBMode-Sel'] = state_sel
        return self._wait(
            'SOFBMode-Sts', state_sts, timeout=timeout)



"""PSSOFB class."""
from copy import deepcopy as _dcopy
from collections.abc import Iterable
import logging as _log
import multiprocessing as _mp
from multiprocessing import sharedctypes as _shm

import numpy as _np
from epics import get_pv as _get_pv
from socket import timeout as _socket_timeout

from ..thread import AsyncWorker as _AsyncWorker
from ..search import PSSearch as _PSSearch
from ..search import IDSearch as _IDSearch
from ..bsmp import SerialError as _SerialError
from ..bsmp import constants as _const_bsmp
from ..devices import StrengthConv as _StrengthConv
from ..epics import CAProcessSpawn as _Process

from .bsmp.constants import ConstFBP as _const_fbp
from .bsmp.constants import UDC_MAX_NR_DEV as _UDC_MAX_NR_DEV
from .bsmp.commands import FBP as _FBP
from .csdev import PSSOFB_MAX_NR_UDC as _PSSOFB_MAX_NR_UDC
from .pructrl.pru import PRU as _PRU
from .pructrl.udc import UDC as _UDC
from .psctrl.pscstatus import PSCStatus as _PSCStatus


class _BBBThread(_AsyncWorker):
    """Class to run methods from a given BBB."""

    def configure_new_run(self, target, args=None):
        """Configure a new run of the thread."""
        args_ = (self.name, )
        if isinstance(args, tuple):
            args_ = args_ + args
        return super().configure_new_run(target, args_)


class PSNamesSOFB:
    """."""

    _sofb = dict()
    _sofb_factory = None

    @staticmethod
    def get_psnames_ch(acc):
        """Return horizontal corrector psnames of a given sector/ID."""
        if 'ID-' in acc:
            return _IDSearch.conv_idname_2_idff_chnames(acc)
        if PSNamesSOFB._sofb_factory is None:
            from ..sofb.csdev import SOFBFactory
            PSNamesSOFB._sofb_factory = SOFBFactory
        if acc not in PSNamesSOFB._sofb:
            PSNamesSOFB._sofb[acc] = PSNamesSOFB._sofb_factory.create(acc)
        return PSNamesSOFB._sofb[acc].ch_names

    @staticmethod
    def get_psnames_cv(acc):
        """Return vertical corrector psnames of a given sector/ID."""
        if 'ID-' in acc:
            return _IDSearch.conv_idname_2_idff_cvnames(acc)
        if PSNamesSOFB._sofb_factory is None:
            from ..sofb.csdev import SOFBFactory
            PSNamesSOFB._sofb_factory = SOFBFactory
        if acc not in PSNamesSOFB._sofb:
            PSNamesSOFB._sofb[acc] = PSNamesSOFB._sofb_factory.create(acc)
        return PSNamesSOFB._sofb[acc].cv_names

    @staticmethod
    def get_psnames_qs(acc):
        """Return skew corrector psnames of a ID."""
        if 'ID-' in acc:
            return _IDSearch.conv_idname_2_idff_qsnames(acc)
        return []

    @staticmethod
    def get_bbbnames(acc):
        """Return bbbnames."""
        corrnames = \
            PSNamesSOFB.get_psnames_ch(acc) + \
            PSNamesSOFB.get_psnames_cv(acc) + \
            PSNamesSOFB.get_psnames_qs(acc)
        bbbnames = set()
        for corrname in corrnames:
            bbbname = _PSSearch.conv_psname_2_bbbname(corrname)
            bbbnames.add(bbbname)
        return list(bbbnames)


class UnitConverter:
    """."""

    DIPOLE_PROPTY = 'Ref-Mon'

    def __init__(self, psnames, dipoleoff=False):
        """."""
        self.psnames = psnames
        self._dipoleoff = dipoleoff
        self._pstype_2_index, self._pstype_2_sconv = self._init_strenconv()

    def _init_strenconv(self):
        # 1. create pstype to StrengthConv dictionary.
        # 2. create pstype to corrector index dictionnary.
        pstype_2_index = dict()
        pstype_2_sconv = dict()
        psnames = self.psnames
        for i, psname in enumerate(psnames):
            pstype = _PSSearch.conv_psname_2_pstype(psname)
            if pstype not in pstype_2_index:
                pstype_2_index[pstype] = []
            pstype_2_index[pstype].append(i)
            if pstype not in pstype_2_sconv:
                # sconv = _NormFact.create(psname.replace(':PS', ':MA'))
                sconv = _StrengthConv(
                    psname, UnitConverter.DIPOLE_PROPTY, auto_monitor_mon=True)
                pstype_2_sconv[pstype] = sconv

        # convert index to numpy array
        for pstype in pstype_2_index:
            pstype_2_index[pstype] = _np.asarray(pstype_2_index[pstype])

        # wait for connection
        for sconv in pstype_2_sconv.values():
            sconv.wait_for_connection()

        return pstype_2_index, pstype_2_sconv

    def conv_curr2stren(self, current):
        """."""
        # benchmarks:
        # 0.163 ms (normalizer)
        # 0.277 ms (sconv)
        strength = _np.zeros(len(current))
        for pstype, index in self._pstype_2_index.items():
            sconv = self._pstype_2_sconv[pstype]
            value = current[index]
            if self._dipoleoff:
                stren = sconv.conv_current_2_strength(
                    currents=value, strengths_dipole=3.0)
            else:
                stren = sconv.conv_current_2_strength(currents=value)

            strength[index] = stren
        return strength

    def conv_stren2curr(self, strength):
        """."""
        # benchmarks:
        # 0.112 ms (normalizer)
        # 0.189 ms (sconv)
        current = _np.full(len(strength), _np.nan, dtype=float)
        for pstype, index in self._pstype_2_index.items():
            sconv = self._pstype_2_sconv[pstype]
            value = strength[index]
            if self._dipoleoff:
                curr = sconv.conv_strength_2_current(
                    strengths=value, strengths_dipole=3.0)
            else:
                curr = sconv.conv_strength_2_current(strengths=value)
            current[index] = curr
        return current


class PSConnSOFB:
    """."""

    MAX_NR_DEVS = _PSSOFB_MAX_NR_UDC * _UDC_MAX_NR_DEV

    PS_PWRSTATE = _PSCStatus.PWRSTATE
    PS_OPMODE = _PSCStatus.OPMODE
    SOCKET_TIMEOUT_ERR = 255
    SERIAL_ERR = 254

    def __init__(
            self, ethbridgeclnt_class, bbbnames=None, mproc=None,
            sofb_update_iocs=False, dipoleoff=False, acc=None):
        """."""
        # check arguments
        if mproc is not None and \
                (not isinstance(mproc, dict) or
                 set(mproc.keys()) != {'rbref', 'ref', 'fret'}):
            raise ValueError('Invalid mproc dictionary!')

        self._dipoleoff = dipoleoff
        self._acc = acc or 'SI'
        self._pru = None
        self._udc = None
        self.bbbnames = bbbnames or _dcopy(PSSOFB.BBBNAMES)
        self.bbb2devs = dict()

        # flag that controls sending sofbupdate signal to IOCs
        self._sofb_update_iocs = sofb_update_iocs

        self._sofb_psnames = \
            PSNamesSOFB.get_psnames_ch(self._acc) + \
            PSNamesSOFB.get_psnames_cv(self._acc) + \
            PSNamesSOFB.get_psnames_qs(self._acc)

        # snapshot of sofb current values
        ncorrs = len(self._sofb_psnames)
        arr = _np.zeros(ncorrs, dtype=float)
        self._sofb_current_rb = arr.copy()
        self._sofb_current_mon = arr.copy()

        if mproc:
            self._sofb_current_readback_ref = mproc['rbref']
            self._sofb_current_refmon = mproc['ref']
            self._sofb_func_return = mproc['fret']
        else:
            self._sofb_current_readback_ref = arr.copy()
            self._sofb_current_refmon = arr.copy()
            self._sofb_func_return = _np.zeros(ncorrs, dtype=int)

        # create sofb and bsmp indices
        self.indcs_bsmp, self.indcs_sofb = self._create_indices()

        # initialize connector objects for bsmp communications.
        self._pru, self._udc = self._init_connectors(ethbridgeclnt_class)

        # dictionaries with comm. ack, state snapshot of all correctors
        # states and threads
        self._dev_ack, self._dev_state, self._threads, self._pvobjs = \
            self._init_threads_dev_state()

        # power supply status objects
        self._pscstatus = [_PSCStatus() for _ in self._sofb_psnames]

        # strength to current converters
        if mproc is None:
            self.converter = UnitConverter(
                self._sofb_psnames, dipoleoff=dipoleoff)

    @property
    def pru(self):
        """Return Beagle-name to PRU-object dictionary."""
        return self._pru

    @property
    def udc(self):
        """Return Beagle-name to UDC-object dictionary."""
        return self._udc

    # --- threads manipulation methods ---

    def threads_shutdown(self):
        """Stop BBB threads."""
        for thr in self._threads:
            thr.stop()
        for thr in self._threads:
            thr.join()

    # --- bsmp methods: invoke communications with correctors ---

    def bsmp_sofb_current_set(self, current):
        """Send current sofb setpoint to power supplies."""
        self._parallel_execution(self._bsmp_current_setpoint, (current, ))

    def bsmp_sofb_kick_set(self, kick):
        """Send kick sofb setpoint to power supplies."""
        current = self.converter.conv_stren2curr(kick)
        self.bsmp_sofb_current_set(current)
        return current

    def bsmp_sofb_current_set_update(self, current):
        """Send current sofb setpoint to power supplies and update."""
        self._parallel_execution(self._bsmp_current_setpoint_update, current)

    def bsmp_sofb_kick_set_update(self, kick):
        """Send kick sofb setpoint to power supplies and update."""
        current = self.converter.conv_stren2curr(kick)
        self.bsmp_sofb_current_set_update(current)
        return current

    def bsmp_update_sofb(self):
        """Update SOFB currents by sending correctors bsmp requests."""
        self._parallel_execution(self._bsmp_update_current)

    def bsmp_update_state(self):
        """Receive from correctors variables and update object attributes."""
        self._parallel_execution(self._bsmp_update_state)

    def bsmp_pwrstate_on(self):
        """Turn all correctors on and then update state."""
        args = (_const_fbp.F_TURN_ON, )
        self._parallel_execution(self._bsmp_execute_function, args)
        self._parallel_execution(self._bsmp_update_state)

    def bsmp_pwrstate_off(self):
        """Turn all correctors off and then update state."""
        args = (_const_fbp.F_TURN_OFF, )
        self._parallel_execution(self._bsmp_execute_function, args)
        self._parallel_execution(self._bsmp_update_state)

    def bsmp_slowref(self):
        """Set all correctors to SlowRef opmode and then update state."""
        args = (_const_fbp.F_SELECT_OP_MODE, _PSCStatus.STATES.SlowRef)
        self._parallel_execution(self._bsmp_execute_function, args)
        self._parallel_execution(self._bsmp_update_state)

    def bsmp_slowrefsync(self):
        """Turn all correctors ro SlowRefSync opmode and then update state."""
        args = (_const_fbp.F_SELECT_OP_MODE, _PSCStatus.STATES.SlowRefSync)
        self._parallel_execution(self._bsmp_execute_function, args)
        self._parallel_execution(self._bsmp_update_state)

    # --- sofb-vector object attribute methods (no comm.) ---

    @property
    def sofb_psnames(self):
        """Return corrector names in SOFB order."""
        return self._sofb_psnames

    @property
    def sofb_kick_refmon(self):
        """Return SOFB kick Ref-Mon vector, as last updated."""
        current = self.sofb_current_refmon
        strength = self.converter.conv_curr2stren(current)
        return strength

    @property
    def sofb_kick_readback_ref(self):
        """Return SOFB kick from current_readback_ref from last setpoint."""
        current = self.sofb_current_readback_ref
        strength = self.converter.conv_curr2stren(current)
        return strength

    @property
    def sofb_current_rb(self):
        """Return SOFB current RB vector, as last updated."""
        return self._sofb_current_rb

    @property
    def sofb_current_refmon(self):
        """Return SOFB current Ref-Mon vector, as last updated."""
        return self._sofb_current_refmon

    @property
    def sofb_current_mon(self):
        """Return SOFB current -Mon vector, as last updated."""
        return self._sofb_current_mon

    @property
    def sofb_current_readback_ref(self):
        """Return SOFB current readback after last setpoint."""
        return self._sofb_current_readback_ref.copy()

    @property
    def sofb_func_return(self):
        """Return function return from last SOFB communication."""
        return self._sofb_func_return.copy()

    def sofb_state_variable_get(self, var_id, dtype=float):
        """Return SOFB vector with bsmp variable values, as last updated."""
        return self._sofb_state_variable(var_id=var_id, dtype=dtype)

    @property
    def sofb_opmode(self):
        """Return SOFB vector with OpMode-Sts values, as last updated."""
        return self._sofb_pscstatus_get('ioc_opmode')

    @property
    def sofb_pwrstate(self):
        """Return SOFB vector with PwrState-Sts values, as last updated."""
        return self._sofb_pscstatus_get('ioc_pwrstate')

    @property
    def sofb_interlock_hard(self):
        """Return SOFB vector with IntlkHard-Sts values, as last updated."""
        return self._sofb_state_variable(
            _const_fbp.V_PS_HARD_INTERLOCKS, dtype=int)

    @property
    def sofb_interlock_soft(self):
        """Return SOFB vector with IntlkSoft-Sts values, as last updated."""
        return self._sofb_state_variable(
            _const_fbp.V_PS_SOFT_INTERLOCKS, dtype=int)

    @staticmethod
    def sofb_vector_issame(vector1, vector2):
        """Check if two sofb vectors are identical within float32 precision."""
        if len(vector1) != len(vector2):
            return False
        # comparison
        vec1 = _np.asarray(vector1, dtype=_np.float32)
        vec2 = _np.asarray(vector2, dtype=_np.float32)
        return _np.all(vec1 == vec2)

    def sofb_conv_psname_2_index(self, psname):
        """."""
        return self._sofb_psnames.index(psname)

    # --- private methods: bsmp comm in parallel ---

    def _parallel_execution(self, target, args=None):
        """Execute 'method' in parallel."""
        # run threads
        for thr in self._threads:
            if not thr.configure_new_run(target, args=args):
                raise RuntimeError('Thread is not ready for job!')
        # wait for run to finish
        for thr in self._threads:
            thr.wait_ready()

    def _bsmp_update_current(self, bbbname):
        """Update SOFB parameters of a single beaglebone."""
        udc = self._udc[bbbname]

        # --- bsmp communication ---
        udc.sofb_update()

        # update currents
        self._update_currents(bbbname)

    def _bsmp_current_setpoint(self, bbbname, curr_sp):
        """Set currents for a single beaglebone."""
        udc = self._udc[bbbname]

        # get indices
        indcs_sofb = self.indcs_sofb[bbbname]
        indcs_bsmp = self.indcs_bsmp[bbbname]

        # initialize setpoint
        # NOTE: curr_sp may contain NaNs. They will be handled by low level
        #    classes.
        setpoint = _np.full(PSConnSOFB.MAX_NR_DEVS, _np.nan)
        setpoint[indcs_bsmp] = curr_sp[indcs_sofb]

        # --- bsmp communication ---
        try:
            udc.sofb_current_set(setpoint)
            # update sofb_func_return
            func_return = udc.sofb_func_return_get()
            self._sofb_func_return[indcs_sofb] = func_return[indcs_bsmp]
        except _socket_timeout:
            # update sofb_func_return indicating socket timeout
            self._sofb_func_return[indcs_sofb] = PSConnSOFB.SOCKET_TIMEOUT_ERR
        except _SerialError:
            self._sofb_func_return[indcs_sofb] = PSConnSOFB.SERIAL_ERR

        # update sofb_current_readback_ref
        current = udc.sofb_current_readback_ref_get()
        if current is None:
            return
        self._sofb_current_readback_ref[indcs_sofb] = current[indcs_bsmp]

        # send signal to IOC to update one power supply state
        if self._sofb_update_iocs:
            pvobj = self._pvobjs[bbbname]
            if pvobj.connected and not pvobj.status:
                pvobj.put(1, wait=False)  # send signal to IOC

    def _bsmp_current_setpoint_update(self, bbbname, curr_sp):
        """."""
        self._bsmp_current_setpoint(bbbname, curr_sp)
        self._bsmp_update_current(bbbname)

    def _bsmp_update_state(self, bbbname):
        udc = self._udc[bbbname]
        devices = self.bbb2devs[bbbname]
        group_id = _const_fbp.G_ALL
        self._dev_state[bbbname] = dict()

        # set dev_ack to False to all devices
        dev_ack = self._dev_ack[bbbname]
        for _, dev_id in devices:
            dev_ack[dev_id] = False

        # execute function for devices
        try:
            for _, dev_id in devices:
                resp = udc[dev_id].read_group_of_variables(group_id=group_id)
                ack, data = resp
                # check anomalous response
                if ack != _const_bsmp.ACK_OK:
                    print(('PSSOFB: could not read '
                           'group of variables id:{}').format(group_id))
                else:
                    self._dev_state[bbbname][dev_id] = data
                    dev_ack[dev_id] = True
        except _SerialError:
            # no communication
            print('PSSOFB: SerialError in bsmp_state_update!')

    def _bsmp_execute_function(self, bbbname, *args):
        """."""
        udc = self._udc[bbbname]
        devices = self.bbb2devs[bbbname]
        function_id, *args = args

        # transform empty tuple in None
        args = args if args else None

        # set dev_ack to False to all devices
        dev_ack = self._dev_ack[bbbname]
        for _, dev_id in devices:
            dev_ack[dev_id] = False

        # execute function for devices
        try:
            for _, dev_id in devices:
                resp = udc[dev_id].execute_function(function_id, args)
                ack, data = resp
                # check anomalous response
                if ack != _const_bsmp.ACK_OK:
                    if isinstance(data, str):
                        data = ord(data)
                    print('PSSOFB: anomalous response !')
                    udc[dev_id].anomalous_response(
                        _const_bsmp.CMD_EXECUTE_FUNCTION, ack,
                        device_id=dev_id,
                        function_id=function_id,
                        data_len=len(data),
                        data_type=type(data),
                        data=data)
                else:
                    dev_ack[dev_id] = True
        except _SerialError:
            # no communication
            return None
        except TypeError:
            # anomalous response triggered unexpected code error
            print('--- debug ----')
            print('devices     : ', devices)
            print('dev_id      : ', dev_id)
            print('function_id : ', function_id)
            print('resp        : ', resp)
            print('data        : ', data)
            raise

    # --- private methods: class initializations. ---

    def _create_indices(self):
        """."""
        self._update_bbb2devs()
        indcs_bsmp = dict()
        indcs_sofb = dict()
        for bbbname in self.bbbnames:
            devs = self.bbb2devs[bbbname]
            # add indcs_udc to dictionary
            indcs_bsmp_ = list()
            indcs_sofb_ = list()
            for psname, bsmpid in devs:
                if psname in self._sofb_psnames:
                    indcs_bsmp_.append(bsmpid-1)
                    indcs_sofb_.append(self._sofb_psnames.index(psname))
            indcs_bsmp[bbbname] = _np.array(indcs_bsmp_)
            indcs_sofb[bbbname] = _np.array(indcs_sofb_)

        return indcs_bsmp, indcs_sofb

    def _init_connectors(self, ethbridgeclnt_class):
        """."""
        pru, udc = self._create_pru_udc(ethbridgeclnt_class)
        self._add_groups_of_variables(udc)

        return pru, udc

    def _add_groups_of_variables(self, udc):
        # add groups of variaable with id>2 to entities class, as
        # required for message-processing verification class objects
        # to work.
        for udc_ in udc.values():
            udc_.add_groups_of_variables()

    def _update_currents(self, bbbname):
        """."""
        udc = self._udc[bbbname]

        # get indices
        indcs_sofb = self.indcs_sofb[bbbname]
        indcs_bsmp = self.indcs_bsmp[bbbname]

        # update -RB
        current = udc.sofb_current_rb_get()
        if current is None:
            return
        self._sofb_current_rb[indcs_sofb] = current[indcs_bsmp]

        # update Ref-Mon
        current = udc.sofb_current_refmon_get()
        self._sofb_current_refmon[indcs_sofb] = current[indcs_bsmp]

        # update -Mon
        current = udc.sofb_current_mon_get()
        self._sofb_current_mon[indcs_sofb] = current[indcs_bsmp]

    def _sofb_state_variable(self, var_id, dtype=int):
        values = _np.zeros(len(self._sofb_psnames), dtype=dtype)
        for bbbname in self.bbbnames:
            # beaglebone dev state
            bbbstate = self._dev_state[bbbname]
            if not bbbstate:
                raise KeyError('PSSOFB.bsmp_state_update not invoked first!')
            # get state values
            bbbvalues = _np.zeros(PSConnSOFB.MAX_NR_DEVS, dtype=dtype)
            for _, dev_id in self.bbb2devs[bbbname]:
                devstate = bbbstate[dev_id]
                bbbvalues[dev_id-1] = devstate[var_id]
            # get indices
            indcs_sofb = self.indcs_sofb[bbbname]
            indcs_bsmp = self.indcs_bsmp[bbbname]
            values[indcs_sofb] = bbbvalues[indcs_bsmp]
        return values

    def _sofb_pscstatus_get(self, attr):
        status = self._sofb_state_variable(
            var_id=_const_fbp.V_PS_STATUS, dtype=int)
        values = _np.zeros(len(self._sofb_psnames), dtype=int)
        for i in range(len(self._sofb_psnames)):
            self._pscstatus[i].ps_status = status[i]
            values[i] = getattr(self._pscstatus[i], attr)
        return values

    def _init_threads_dev_state(self):
        dev_ack = dict()  # last ack state of bsmp comm.
        dev_state = dict()  # snapshot of device variable values
        threads = list()  # threads to run BBB methods
        pvobjs = dict()  # SOFBUpdate PVs
        for bbbname, bsmpdevs in self.bbb2devs.items():
            dev_ack[bbbname] = {bsmp[1]: True for bsmp in bsmpdevs}
            dev_state[bbbname] = {bsmp[1]: dict() for bsmp in bsmpdevs}
            if self._sofb_update_iocs:
                pvname = bsmpdevs[0][0] + ':SOFBUpdate-Cmd'
                pvobjs[bbbname] = _get_pv(pvname)
            thread = _BBBThread(name=bbbname)
            thread.start()
            threads.append(thread)
        return dev_ack, dev_state, threads, pvobjs

    def _update_bbb2devs(self):
        """."""
        for bbbname in self.bbbnames:
            devs = _PSSearch.conv_bbbname_2_bsmps(bbbname)
            self.bbb2devs[bbbname] = devs

    def _create_pru_udc(self, ethbridgeclnt_class):
        """."""
        pru, udc = dict(), dict()

        # signal FBP comm class not to print comm. error statements
        _FBP.SOFB_PRINT_COMM_ERRORS = False

        for bbbname, bsmpdevs in self.bbb2devs.items():

            # create PRU object for bsmp communication
            pru_ = _PRU(ethbridgeclnt_class, bbbname)

            # for a given beagle, gather devices ids
            device_ids = [bsmp[1] for bsmp in bsmpdevs]

            # create UDC object to access power supply bsmp comm. methods
            udc_ = _UDC(pru=pru_, psmodel='FBP', device_ids=device_ids)

            # add PRU and UDC objects to disctionaries.
            pru[bbbname] = pru_
            udc[bbbname] = udc_

        return pru, udc


class PSSOFB:
    """."""

    BBBNAMES = (
        'IA-01RaCtrl:CO-PSCtrl-SI2',
        'IA-02RaCtrl:CO-PSCtrl-SI2',
        'IA-03RaCtrl:CO-PSCtrl-SI2',
        'IA-04RaCtrl:CO-PSCtrl-SI2',
        'IA-05RaCtrl:CO-PSCtrl-SI2',
        'IA-06RaCtrl:CO-PSCtrl-SI2',
        'IA-07RaCtrl:CO-PSCtrl-SI2',
        'IA-08RaCtrl:CO-PSCtrl-SI2',
        'IA-09RaCtrl:CO-PSCtrl-SI2',
        'IA-10RaCtrl:CO-PSCtrl-SI2',
        'IA-11RaCtrl:CO-PSCtrl-SI2',
        'IA-12RaCtrl:CO-PSCtrl-SI2',
        'IA-13RaCtrl:CO-PSCtrl-SI2',
        'IA-14RaCtrl:CO-PSCtrl-SI2',
        'IA-15RaCtrl:CO-PSCtrl-SI2',
        'IA-16RaCtrl:CO-PSCtrl-SI2',
        'IA-17RaCtrl:CO-PSCtrl-SI2',
        'IA-18RaCtrl:CO-PSCtrl-SI2',
        'IA-19RaCtrl:CO-PSCtrl-SI2',
        'IA-20RaCtrl:CO-PSCtrl-SI2',
        'IA-01RaCtrl:CO-PSCtrl-SI4',
        'IA-02RaCtrl:CO-PSCtrl-SI4',
        'IA-03RaCtrl:CO-PSCtrl-SI4',
        'IA-04RaCtrl:CO-PSCtrl-SI4',
        'IA-05RaCtrl:CO-PSCtrl-SI4',
        'IA-06RaCtrl:CO-PSCtrl-SI4',
        'IA-07RaCtrl:CO-PSCtrl-SI4',
        'IA-08RaCtrl:CO-PSCtrl-SI4',
        'IA-09RaCtrl:CO-PSCtrl-SI4',
        'IA-10RaCtrl:CO-PSCtrl-SI4',
        'IA-11RaCtrl:CO-PSCtrl-SI4',
        'IA-12RaCtrl:CO-PSCtrl-SI4',
        'IA-13RaCtrl:CO-PSCtrl-SI4',
        'IA-14RaCtrl:CO-PSCtrl-SI4',
        'IA-15RaCtrl:CO-PSCtrl-SI4',
        'IA-16RaCtrl:CO-PSCtrl-SI4',
        'IA-17RaCtrl:CO-PSCtrl-SI4',
        'IA-18RaCtrl:CO-PSCtrl-SI4',
        'IA-19RaCtrl:CO-PSCtrl-SI4',
        'IA-20RaCtrl:CO-PSCtrl-SI4',
        )
    BBB2DEVS = dict()

    def __init__(
            self, ethbridgeclnt_class, nr_procs=8, asynchronous=False,
            sofb_update_iocs=False, dipoleoff=False):
        """."""
        self._dipoleoff = dipoleoff
        self._acc = 'SI'
        self._async = asynchronous
        self._sofb_update_iocs = sofb_update_iocs

        self._sofb_psnames = \
            PSNamesSOFB.get_psnames_ch(self._acc) + \
            PSNamesSOFB.get_psnames_cv(self._acc)

        # snapshot of sofb current values
        ncorrs = len(self._sofb_psnames)
        arr = _np.zeros(ncorrs, dtype=float)
        self._sofb_current_readback_ref = arr.copy()
        self._sofb_current_refmon = arr.copy()
        self._sofb_func_return = _np.zeros(ncorrs, dtype=int)

        # Unit converter.
        self.converter = UnitConverter(
            self._sofb_psnames, dipoleoff=dipoleoff)

        # Worker Processes control and synchronization
        self._ethbridge_cls = ethbridgeclnt_class
        self._nr_procs = nr_procs
        self._doneevts = []
        self._procs = []
        self._pipes = []

    # --- General class properties ---

    @property
    def asynchronous(self):
        """."""
        return self._async

    @asynchronous.setter
    def asynchronous(self, val):
        self._async = bool(val)

    def wait(self, timeout=None):
        """."""
        for i, doneevt in enumerate(self._doneevts):
            if not doneevt.wait(timeout=timeout):
                _log.error('Wait Done timed out for process '+str(i))
                return False
        return True

    def is_ready(self):
        """."""
        for i, doneevt in enumerate(self._doneevts):
            if not doneevt.is_set():
                _log.error('Ready: not done for process '+str(i))
                return False
        return True

    # --- processes manipulation methods ---

    def processes_start(self):
        """."""
        # get the start method of the Processes that will be launched:
        spw = _mp.get_context('spawn')

        # Create shared memory objects to be shared with worker processes.
        arr = self._sofb_current_readback_ref

        rbref = _shm.Array(_shm.ctypes.c_double, arr.size, lock=False)
        self._sofb_current_readback_ref = _np.ndarray(
            arr.shape, dtype=arr.dtype, buffer=memoryview(rbref))

        ref = _shm.Array(_shm.ctypes.c_double, arr.size, lock=False)
        self._sofb_current_refmon = _np.ndarray(
            arr.shape, dtype=arr.dtype, buffer=memoryview(ref))

        fret = _shm.Array(_shm.ctypes.c_int, arr.size, lock=False)
        self._sofb_func_return = _np.ndarray(
            arr.shape, dtype=_np.int32, buffer=memoryview(fret))

        # Unit converter.
        self.converter = UnitConverter(
            self._sofb_psnames, dipoleoff=self._dipoleoff)

        # subdivide the pv list for the processes
        nr_bbbs = len(PSSOFB.BBBNAMES)
        div = nr_bbbs // self._nr_procs
        rem = nr_bbbs % self._nr_procs
        sub = [div*i + min(i, rem) for i in range(self._nr_procs+1)]
        for i in range(self._nr_procs):
            bbbnames = PSSOFB.BBBNAMES[sub[i]:sub[i+1]]
            # NOTE: It is crucial to use the Event class from the appropriate
            # context, otherwise it will fail for 'spawn' start method.
            doneevt = spw.Event()
            doneevt.set()
            theirs, mine = spw.Pipe(duplex=False)
            proc = _Process(
                target=PSSOFB._run_process,
                args=(self._ethbridge_cls, bbbnames, theirs, doneevt,
                      arr.shape, rbref, ref, fret, self._sofb_update_iocs,
                      self._dipoleoff),
                daemon=True)
            proc.start()
            self._procs.append(proc)
            self._doneevts.append(doneevt)
            self._pipes.append(mine)

    def processes_shutdown(self):
        """."""
        for pipe in self._pipes:
            pipe.send(None)
            pipe.close()
        for proc in self._procs:
            proc.join()

    # --- bspm communication methods ---

    def bsmp_sofb_current_set(self, current):
        """Send current sofb setpoint to power supplies."""
        self._parallel_execution('bsmp_sofb_current_set', (current, ))

    def bsmp_sofb_kick_set(self, kick):
        """Send kick sofb setpoint to power supplies."""
        current = self.converter.conv_stren2curr(kick)
        self.bsmp_sofb_current_set(current)

    def bsmp_sofb_current_set_update(self, current):
        """Send current sofb setpoint to power supplies and update."""
        self._parallel_execution('bsmp_sofb_current_set_update', (current, ))

    def bsmp_sofb_kick_set_update(self, kick):
        """Send kick sofb setpoint to power supplies and update."""
        current = self.converter.conv_stren2curr(kick)
        self.bsmp_sofb_current_set_update(current)

    def bsmp_update_sofb(self):
        """Update SOFB currents by sending correctors bsmp requests."""
        self._parallel_execution('bsmp_update_sofb')

    def bsmp_update_state(self):
        """Receive from correctors variables and update object attributes."""
        self._parallel_execution('bsmp_update_state')

    def bsmp_pwrstate_on(self):
        """Turn all correctors on and then update state."""
        self._parallel_execution('bsmp_pwrstate_on')

    def bsmp_pwrstate_off(self):
        """Turn all correctors off and then update state."""
        self._parallel_execution('bsmp_pwrstate_off')

    def bsmp_slowref(self):
        """Set all correctors to SlowRef opmode and then update state."""
        self._parallel_execution('bsmp_slowref')

    def bsmp_slowrefsync(self):
        """Turn all correctors ro SlowRefSync opmode and then update state."""
        self._parallel_execution('bsmp_slowrefsync')

    # --- sofb-vector object attribute methods (no comm.) ---

    @property
    def sofb_psnames(self):
        """Return corrector names in SOFB order."""
        return self._sofb_psnames

    @property
    def sofb_current_refmon(self):
        """Return SOFB current readback after last setpoint."""
        return self._sofb_current_refmon.copy()

    @property
    def sofb_current_readback_ref(self):
        """Return SOFB current readback after last setpoint."""
        return self._sofb_current_readback_ref.copy()

    @property
    def sofb_kick_refmon(self):
        """Return SOFB kick from current_readback_ref from last setpoint."""
        return self.converter.conv_curr2stren(self.sofb_current_refmon)

    @property
    def sofb_kick_readback_ref(self):
        """Return SOFB kick from current_readback_ref from last setpoint."""
        return self.converter.conv_curr2stren(self.sofb_current_readback_ref)

    @property
    def sofb_func_return(self):
        """Return SOFB function return from last communication."""
        return self._sofb_func_return.copy()

    @staticmethod
    def sofb_vector_issame(vector1, vector2):
        """Check if two sofb vectors are identical within float32 precision."""
        if len(vector1) != len(vector2):
            return False
        # comparison
        vec1 = _np.asarray(vector1, dtype=_np.float32)
        vec2 = _np.asarray(vector2, dtype=_np.float32)
        return _np.all(vec1 == vec2)

    def sofb_conv_psname_2_index(self, psname):
        """."""
        return self._sofb_psnames.index(psname)

    @staticmethod
    def _run_process(
            ethbridgeclnt_class, bbbnames, pipe, doneevt,
            shape, rbref, ref, fret, sofb_update_iocs, dipoleoff):
        """."""
        mproc = {
            'rbref': _np.ndarray(shape, dtype=float, buffer=memoryview(rbref)),
            'ref': _np.ndarray(shape, dtype=float, buffer=memoryview(ref)),
            'fret': _np.ndarray(
                shape, dtype=_np.int32, buffer=memoryview(fret))}
        psconnsofb = PSConnSOFB(
            ethbridgeclnt_class, bbbnames, mproc=mproc,
            sofb_update_iocs=sofb_update_iocs, dipoleoff=dipoleoff)

        while True:
            rec = pipe.recv()
            if not rec:
                break
            meth, args = rec
            if isinstance(meth, str):
                if isinstance(args, Iterable):
                    getattr(psconnsofb, meth)(*args)
                else:
                    getattr(psconnsofb, meth)()
            doneevt.set()
        psconnsofb.threads_shutdown()
        pipe.close()

    # --- private methods: get properties ---

    def _parallel_execution(self, target_name, args=None):
        """Execute 'method' in parallel."""
        for pipe, doneevt in zip(self._pipes, self._doneevts):
            doneevt.clear()
            pipe.send((target_name, args))

        if not self._async:
            self.wait()



"""Beam stability info App."""

import os as _os
import logging as _log
import time as _time
from functools import partial as _part
import epics as _epics

import numpy as _np

from ..epics import PV as _PV
from ..callbacks import Callback as _Callback
from ..envars import VACA_PREFIX as _vaca_prefix
from ..namesys import SiriusPVName as _PVName
from ..devices import Device as _Device

from .csdev import StabInfoConst as _Const


class App(_Callback):
    """Main application for handling machine shift."""

    SCAN_FREQUENCY = 2  # [Hz]

    def __init__(self):
        """Class constructor."""
        super().__init__()
        self._const = _Const()
        self._pvs_database = self._const.get_propty_database()

        self._bbbh_oscamp = 0
        self._bbbv_oscamp = 0
        self._bbbl_oscamp = 0
        self._bbbh_oscamp_thres = self._const.BBBH_OSCAMP_THRES
        self._bbbv_oscamp_thres = self._const.BBBV_OSCAMP_THRES
        self._bbbl_oscamp_thres = self._const.BBBL_OSCAMP_THRES
        self._bbbh_status = self._const.StabUnstab.Stable
        self._bbbv_status = self._const.StabUnstab.Stable
        self._bbbl_status = self._const.StabUnstab.Stable
        self._bbbh_calibfactor = self._const.DEF_BBBH_CALIBFACTOR
        self._bbbv_calibfactor = self._const.DEF_BBBV_CALIBFACTOR
        self._bbbl_calibfactor = self._const.DEF_BBBL_CALIBFACTOR
        self.read_autosave_file()

        # connections
        # use pyepics recommendations for threading
        _epics.ca.use_initial_context()

        self._dev_bbb = dict()
        bbbprops = ['SRAM_PEAK1', ]
        for plane in ['H', 'V', 'L']:
            devname = 'SI-Glob:DI-BbBProc-'+plane
            self._dev_bbb[plane] = _Device(devname, props2init=bbbprops)

        sicurrent_pvname = _PVName(
            'SI-Glob:AP-CurrInfo:Current-Mon').substitute(
                prefix=_vaca_prefix)
        self.si_current_pv = _PV(sicurrent_pvname, connection_timeout=0.05)

        # pvs to write methods
        self.map_pv2write = {
            'BbBHCalibFactor-SP': _part(self.set_bbbcalibfactor, 'h'),
            'BbBVCalibFactor-SP': _part(self.set_bbbcalibfactor, 'v'),
            'BbBLCalibFactor-SP': _part(self.set_bbbcalibfactor, 'l'),
        }

        self.quit = False
        self.scanning = False
        self._thread_update = _epics.ca.CAThread(
            target=self._update_pvs, daemon=True)
        self._thread_update.start()

    def init_database(self):
        """Set initial PV values."""
        self.run_callbacks('BbBHCalibFactor-SP', self._bbbh_calibfactor)
        self.run_callbacks('BbBHCalibFactor-RB', self._bbbh_calibfactor)
        self.run_callbacks('BbBHOscAmp-Mon', self._bbbh_oscamp)
        self.run_callbacks('BbBHOscAmpThres-Cte', self._bbbh_oscamp_thres)
        self.run_callbacks('BbBHStatus-Mon', self._bbbh_status)
        self.run_callbacks('BbBVCalibFactor-SP', self._bbbv_calibfactor)
        self.run_callbacks('BbBVCalibFactor-RB', self._bbbv_calibfactor)
        self.run_callbacks('BbBVOscAmp-Mon', self._bbbv_oscamp)
        self.run_callbacks('BbBVOscAmpThres-Cte', self._bbbv_oscamp_thres)
        self.run_callbacks('BbBVStatus-Mon', self._bbbv_status)
        self.run_callbacks('BbBLCalibFactor-SP', self._bbbl_calibfactor)
        self.run_callbacks('BbBLCalibFactor-RB', self._bbbl_calibfactor)
        self.run_callbacks('BbBLOscAmp-Mon', self._bbbl_oscamp)
        self.run_callbacks('BbBLOscAmpThres-Cte', self._bbbl_oscamp_thres)
        self.run_callbacks('BbBLStatus-Mon', self._bbbl_status)
        self._update_log('Started.')

    @property
    def pvs_database(self):
        """Return pvs_database."""
        return self._pvs_database

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def read(self, reason):
        """Read from IOC database."""
        value = None
        return value

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        _log.info('Write received for: %s --> %s', reason, str(value))
        if reason in self.map_pv2write.keys():
            status = self.map_pv2write[reason](value)
            _log.info('%s Write for: %s --> %s',
                      str(status).upper(), reason, str(value))
            return status
        _log.warning('PV %s does not have a set function.', reason)
        return False

    def set_bbbcalibfactor(self, plane, value):
        """Set BbB calibration factor."""
        setattr(self, '_bbb'+plane+'_calibfactor', value)
        self.run_callbacks('BbB'+plane.upper()+'CalibFactor-RB', value)
        self.update_autosave_file()
        return True

    #  ----- log auxiliary methods -----

    def _update_log(self, msg):
        if 'ERR' in msg:
            _log.error(msg[4:])
        elif 'FATAL' in msg:
            _log.error(msg[6:])
        elif 'WARN' in msg:
            _log.warning(msg[5:])
        else:
            _log.info(msg)
        self.run_callbacks('Log-Mon', msg)

    # ----- auto save methods -----

    @property
    def auto_save_data(self):
        """Data to be saved in autosave file."""
        data = [
            self._bbbh_calibfactor,
            self._bbbv_calibfactor,
            self._bbbl_calibfactor,
            ]
        return _np.array(data)

    def read_autosave_file(self):
        """Read autosave file."""
        filename = self._const.autosave_fname
        data = None
        if _os.path.isfile(filename):
            data = _np.loadtxt(filename)
        if data is not None:
            bbbh, bbbv, bbbl = data
            self._bbbh_calibfactor = bbbh
            self._bbbv_calibfactor = bbbv
            self._bbbl_calibfactor = bbbl
        else:
            _log.info(
                'No backup file was found, default values '
                'were used for calibration factors.')
            self.update_autosave_file()
            _log.info('First autosave file was created.')

    def update_autosave_file(self):
        """Update autosave file."""
        path = _os.path.split(self._const.autosave_fname)[0]
        _os.makedirs(path, exist_ok=True)
        _np.savetxt(
            self._const.autosave_fname, self.auto_save_data)

    # ----- update pvs methods -----

    def _update_pvs(self):
        tplanned = 1.0/App.SCAN_FREQUENCY
        while not self.quit:
            if not self.scanning:
                _time.sleep(tplanned)
                continue

            _t0 = _time.time()

            pvcurr = self.si_current_pv
            if pvcurr.connected and pvcurr.value is not None:
                curr = pvcurr.value
                for pln in ['H', 'V', 'L']:
                    dev = self._dev_bbb[pln]
                    harmn = self._const.SI_HARMNUM
                    calib = getattr(self, '_bbb'+pln.lower()+'_calibfactor')
                    thres = getattr(self, '_bbb'+pln.lower()+'_oscamp_thres')
                    cthres = self._const.CURR_THRES
                    oscamp_db = dev['SRAM_PEAK1']
                    if oscamp_db is not None:
                        oscamp_phy = 10**(oscamp_db/20) * harmn / curr / calib
                        self.run_callbacks('BbB'+pln+'OscAmp-Mon', oscamp_phy)
                        status = curr > cthres and oscamp_phy > thres
                        self.run_callbacks('BbB'+pln+'Status-Mon', int(status))

            ttook = _time.time() - _t0
            tsleep = tplanned - ttook
            if tsleep > 0:
                _time.sleep(tsleep)
            else:
                _log.warning(
                    'Update loop took more than planned... '
                    '{0:.3f}/{1:.3f} s'.format(ttook, tplanned))



"""Define the high level classes."""

import time as _time
from functools import partial as _partial, reduce as _reduce
from operator import or_ as _or_, and_ as _and_
import logging as _log
from threading import Lock as _Lock

import numpy as _np

from ..util import mode as _mode
from ..thread import RepeaterThread as _Timer
from ..search import HLTimeSearch as _HLSearch
from ..namesys import SiriusPVName as _PVName
from ..callbacks import Callback as _Callback

from .csdev import get_hl_trigger_database as _get_hl_trigger_database
from .ll_classes import get_ll_trigger as _get_ll_trigger


# HL == High Level
class _BaseHL(_Callback):
    """Define a High Level interface.

    Determine how to connect the driver with the classes which communicate
    with low level IOCs.
    """

    class Alarm:
        """."""

        NO = 0
        STATE = 7
        COMM = 9
        LINK = 14

    class Severity:
        """."""

        NO = 0
        MINOR = 1
        MAJOR = 2
        INVALID = 3

    _SUFFIX_FOR_PROPS = {}

    def __init__(self, prefix, ll_objs, callback=None):
        """Appropriately initialize the instance.

        prefix = first part of the pv name of this object.
        callback = callable to be called when readbacks are updated.
        """
        super().__init__(callback=callback)
        _log.info(prefix + ' Starting.')
        self.prefix = prefix
        self._ll_objs = ll_objs
        self._funs_combine_values = self._define_funs_combine_values()
        self._all_props_suffix = self._get_properties_suffix()
        self._timer = None
        self._start_timer()
        for obj in self._ll_objs:
            obj.add_callback(self._on_change_pvs)

    @property
    def connected(self):
        """."""
        return all(map(lambda x: x.connected, self._ll_objs))

    def process(self):
        """."""
        return

    def wait_for_connection(self, timeout=None):
        """."""
        return all(map(
            lambda x: x.wait_for_connection(timeout=timeout), self._ll_objs))

    @property
    def locked(self):
        """."""
        dic_ = self._combine_default(map(lambda x: x.locked, self._ll_objs))
        return dic_['value']

    @locked.setter
    def locked(self, value):
        for obj in self._ll_objs:
            obj.locked = bool(value)
        self.run_callbacks(self._get_pv_name('LowLvlLock'), value=bool(value))

    def get_database(self):
        """."""
        return dict()

    def write(self, prop_name, value):
        """Write function to be called by the IOC to set high level properties.

        It not only sets the new high level property value but also forwards it
        to the low level classes.
        """
        if value is None:
            return False
        return _reduce(_and_, map(
            lambda x: x.write(prop_name, value), self._ll_objs))

    def read(self, prop_name, is_sp=False):
        """Read."""
        fun = self._funs_combine_values.get(prop_name, self._combine_default)
        vals = [x.read(prop_name, is_sp=is_sp) for x in self._ll_objs]
        return fun(vals)

    def readall(self, is_sp=False):
        """Read all."""
        values = dict()
        for prop, suf in self._all_props_suffix.items():
            if _PVName.is_cmd_pv(suf) or (is_sp and not _PVName.is_rb_pv(suf)):
                continue
            value = self.read(prop, is_sp=is_sp)
            if value is None or not value:
                continue
            values[prop] = value
        return values

    def get_map2writepvs(self):
        """Get the database."""
        dbase = self.get_database()
        map2write = dict()
        for pvname in dbase:
            prop = self._get_prop_name(pvname)
            if _PVName.is_write_pv(pvname):
                map2write[pvname] = _partial(self.write, prop)
        return map2write

    def get_map2readpvs(self):
        """Get the database."""
        dbase = self.get_database()
        map2readpvs = dict()
        for pvname in dbase:
            if _PVName.is_cte_pv(pvname) or _PVName.is_cmd_pv(pvname):
                continue
            prop = self._get_prop_name(pvname)
            map2readpvs[pvname] = _partial(
                self.read, prop, is_sp=_PVName.is_sp_pv(pvname))
        return map2readpvs

    def _on_change_pvs(self, *args, **kwargs):
        _ = args
        _ = kwargs
        self._start_timer()

    def _start_timer(self, *args, **kwargs):
        if self._timer is not None and self._timer.is_alive():
            self._timer.reset()
        else:
            self._timer = _Timer(0, self._update_pvs_thread, niter=1)
            self._timer.start()

    def _update_pvs_thread(self):
        _time.sleep(1/10)  # limit update in 10Hz
        for prop, suf in self._all_props_suffix.items():
            if _PVName.is_cmd_pv(suf):
                continue
            value = self.read(prop)
            if value is None:
                return
            self.run_callbacks(self._get_pv_name(prop), **value)
            # Uncomment this if you want to update setpoint PVs as well
            # This is not recommended though, because it can create very
            # strange unusual behavior with widgets such as spinbox in PyDM
            # and CS-Studio.
            if not _PVName.is_rb_pv(suf):
                continue
            value = self.read(prop, is_sp=True)
            if value is None:
                return
            self.run_callbacks(self._get_pv_name(prop, is_sp=True), **value)

    def _define_funs_combine_values(self):
        """Define a dictionary of functions to combine low level values.

        Any property not defined here will use the default method:
            _combine_default.
        """
        return dict()

    def _get_properties_suffix(self):
        dbase = self.get_database()
        props = dict()
        for pvname in dbase:
            if not _PVName.is_sp_pv(pvname) and not _PVName.is_cte_pv(pvname):
                prop, suf = self._get_prop_name(pvname, with_suffix=True)
                props[prop] = suf
        return props

    def _get_pv_name(self, prop, is_sp=False):
        pvname = self.prefix + prop + self._all_props_suffix[prop]
        if is_sp:
            pvname = _PVName.from_rb2sp(pvname)
        return pvname

    def _get_prop_name(self, pvname, with_suffix=False):
        ret = pvname[len(self.prefix):].split('-')
        if with_suffix:
            return ret[0], '-' + ret[1]
        else:
            return ret[0]

    def _combine_default(self, values):
        dic_ = {
            'alarm': self.Alarm.COMM,
            'severity': self.Severity.INVALID}
        values = [val for val in values if val is not None]
        if not values:
            dic_['value'] = None
            return dic_
        value, cnt = _mode(sorted(values))
        dic_['value'] = value
        if cnt == len(self._ll_objs):
            dic_['alarm'] = self.Alarm.NO
            dic_['severity'] = self.Severity.NO
        return dic_


class HLTrigger(_BaseHL):
    """High level Trigger interface."""

    def __init__(self, hl_trigger, callback=None):
        """Appropriately initialize the instance."""

        src_enums = _get_hl_trigger_database(hl_trigger=hl_trigger)
        src_enums = src_enums['Src-Sel']['enums']
        ll_obj_names = _HLSearch.get_ll_trigger_names(hl_trigger)

        self._hldelay_lock = _Lock()
        self._hldelay = 0
        self._hldeltadelay = _np.zeros(len(ll_obj_names), dtype=int)

        ll_objs = list()
        for name in ll_obj_names:
            ll_objs.append(_get_ll_trigger(
                channel=name, source_enums=src_enums))
        super().__init__(hl_trigger + ':', ll_objs, callback=callback)

    def process(self):
        """."""
        value = self.read('InInjTable')
        if value is None:
            return
        self.run_callbacks(self._get_pv_name('InInjTable'), **value)
        return

    def write(self, prop_name, value):
        """Call to set high level properties.

        It not only sets the new high level property value but also forwards it
        to the low level classes.
        """
        if value is None:
            return False
        if prop_name.startswith('DelayRaw'):
            with self._hldelay_lock:
                self._hldelay = int(value)
                value = self._hldelay + self._hldeltadelay
        elif prop_name.startswith('Delay'):
            prop_name = prop_name.replace('Delay', 'DelayRaw')
            with self._hldelay_lock:
                self._hldelay = int(round(value / self._ll_objs[0].base_del))
                value = self._hldelay + self._hldeltadelay
        elif prop_name.startswith('DeltaDelayRaw'):
            prop_name = prop_name.replace('DeltaDelayRaw', 'DelayRaw')
            with self._hldelay_lock:
                value = _np.array(value, dtype=int, ndmin=1)
                self._update_deltadelay(value)
                value = self._hldelay + self._hldeltadelay
        elif prop_name.startswith('DeltaDelay'):
            prop_name = prop_name.replace('DeltaDelay', 'DelayRaw')
            with self._hldelay_lock:
                value = _np.array(value, ndmin=1)
                value = _np.round(value / self._ll_objs[0].base_del)
                value = _np.array(value, dtype=int)
                self._update_deltadelay(value)
                value = self._hldelay + self._hldeltadelay
        else:
            value = len(self._ll_objs) * [value, ]

        boo = True
        for val, obj in zip(value, self._ll_objs):
            boo &= obj.write(prop_name, val)
        return boo

    def read(self, prop_name, is_sp=False):
        """."""
        fun = self._funs_combine_values.get(prop_name, self._combine_default)
        if prop_name.startswith('DeltaDelay'):
            prop_name = prop_name.replace('DeltaDelay', 'Delay')
        if prop_name.startswith('DeltaDelayRaw'):
            prop_name = prop_name.replace('DeltaDelayRaw', 'DelayRaw')
        if prop_name.startswith('LowLvlLock') and is_sp:
            vals = len(self._ll_objs) * [self.locked, ]
        else:
            vals = [x.read(prop_name, is_sp=is_sp) for x in self._ll_objs]
        return fun(vals)

    def get_database(self):
        """Get the database."""
        return _get_hl_trigger_database(
            hl_trigger=self.prefix[:-1], prefix=self.prefix)

    def get_ll_triggers(self):
        """."""
        return self._ll_objs

    def get_ll_trigger_names(self):
        """."""
        return _HLSearch.get_ll_trigger_names(self.prefix[:-1])

    def get_ll_channels(self):
        """."""
        return _HLSearch.get_hl_trigger_channels(self.prefix[:-1])

    def _update_deltadelay(self, value):
        siz = min(value.size, self._hldeltadelay.size)
        self._hldeltadelay[:siz] = value[:siz]
        mini = self._hldeltadelay.min()
        self._hldelay += mini
        self._hldeltadelay -= mini
        self._hldelay = 0 if self._hldelay <= 0 else self._hldelay

    def _combine_status(self, values):
        status_or = _reduce(_or_, values)
        status_and = _reduce(_and_, values)
        alarm = self.Alarm.NO
        severity = self.Severity.NO
        if status_or != status_and:
            alarm = self.Alarm.COMM
            severity = self.Severity.INVALID
        return {'value': status_or, 'alarm': alarm, 'severity': severity}

    def _combine_deltadelay(self, values):
        if any(map(lambda x: x is None, values)):
            return {
                'value': None,
                'alarm': self.Alarm.COMM,
                'severity': self.Severity.INVALID}
        alarm = self.Alarm.NO
        severity = self.Severity.NO
        values = _np.array(values) - min(values)
        return {'value': values, 'alarm': alarm, 'severity': severity}

    def _combine_delay(self, values):
        if any(map(lambda x: x is None, values)):
            return {
                'value': None,
                'alarm': self.Alarm.COMM,
                'severity': self.Severity.INVALID}
        alarm = self.Alarm.NO
        severity = self.Severity.NO
        values = min(values)
        return {'value': values, 'alarm': alarm, 'severity': severity}

    def _define_funs_combine_values(self):
        """Define a dictionary of functions to combine low level values."""
        return {
            'Status': self._combine_status,
            'DeltaDelay': self._combine_deltadelay,
            'DeltaDelayRaw': self._combine_deltadelay,
            'Delay': self._combine_delay,
            'DelayRaw': self._combine_delay,
            'TotalDelay': self._combine_delay,
            'TotalDelayRaw': self._combine_delay}



#!/usr/bin/env python-sirius

"""Unittest module for currinfo.py."""

from unittest import TestCase
from siriuspy import util
from siriuspy.currinfo import csdev
from siriuspy.currinfo.csdev import \
    get_currinfo_database, \
    get_lifetime_database

PUB_INTERFACE = (
        'ETypes',
        'Const',
        'get_currinfo_database',
        'get_litbts_currinfo_database',
        'get_bo_currinfo_database',
        'get_si_currinfo_database',
        'get_lifetime_database'
    )


class TestCurrInfoCSDevice(TestCase):
    """Test siriuspy.currinfo.csdev module."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(csdev, PUB_INTERFACE)
        self.assertTrue(valid)

    def test_get_currinfo_database(self):
        """Test get_currinfo_database."""

        # ---- SI ----
        dbase = get_currinfo_database('SI')
        self.assertIsInstance(dbase, dict)

        # PV names
        self.assertTrue('SI-Glob:AP-CurrInfo:Version-Cte' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:Current-Mon' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:StoredEBeam-Mon' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:DCCT-Sel' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:DCCT-Sts' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:DCCTFltCheck-Sel' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:DCCTFltCheck-Sts' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:Charge-Mon' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:InjCurr-Mon' in dbase)
        self.assertTrue('SI-Glob:AP-CurrInfo:InjCharge-Mon' in dbase)
        self.assertTrue('AS-Glob:AP-CurrInfo:InjCount-Mon' in dbase)

        # PVs units
        self.assertEqual(
            dbase['SI-Glob:AP-CurrInfo:Current-Mon']['unit'], 'mA')
        self.assertEqual(
            dbase['SI-Glob:AP-CurrInfo:Charge-Mon']['unit'], 'A.h')
        self.assertEqual(
            dbase['SI-Glob:AP-CurrInfo:InjCurr-Mon']['unit'], 'mA')
        self.assertEqual(
            dbase['SI-Glob:AP-CurrInfo:InjCharge-Mon']['unit'], 'nC')

        # ---- BO ----
        dbase = get_currinfo_database('BO')
        self.assertIsInstance(dbase, dict)

        # PV names
        self.assertTrue('RawReadings-Mon' in dbase)
        self.assertTrue('Current150MeV-Mon' in dbase)
        self.assertTrue('Current1GeV-Mon' in dbase)
        self.assertTrue('Current2GeV-Mon' in dbase)
        self.assertTrue('Current3GeV-Mon' in dbase)
        self.assertTrue('Charge150MeV-Mon' in dbase)
        self.assertTrue('Charge1GeV-Mon' in dbase)
        self.assertTrue('Charge2GeV-Mon' in dbase)
        self.assertTrue('Charge3GeV-Mon' in dbase)
        self.assertTrue('RampEff-Mon' in dbase)

        # PVs units
        self.assertEqual(dbase['RawReadings-Mon']['unit'], 'mA')
        self.assertEqual(dbase['Current150MeV-Mon']['unit'], 'mA')
        self.assertEqual(dbase['Current1GeV-Mon']['unit'], 'mA')
        self.assertEqual(dbase['Current2GeV-Mon']['unit'], 'mA')
        self.assertEqual(dbase['Current3GeV-Mon']['unit'], 'mA')
        self.assertEqual(dbase['Charge150MeV-Mon']['unit'], 'nC')
        self.assertEqual(dbase['Charge1GeV-Mon']['unit'], 'nC')
        self.assertEqual(dbase['Charge2GeV-Mon']['unit'], 'nC')
        self.assertEqual(dbase['Charge3GeV-Mon']['unit'], 'nC')
        self.assertEqual(dbase['RampEff-Mon']['unit'], '%')

    def test_get_lifetime_database(self):
        """Test get_lifetime_database."""
        dbase = get_lifetime_database()
        self.assertIsInstance(dbase, dict)

        # PV names
        self.assertTrue('VersionLifetime-Cte' in dbase)
        self.assertTrue('LtFitMode-Sel' in dbase)
        self.assertTrue('LtFitMode-Sts' in dbase)
        self.assertTrue('MaxSplIntvl-SP' in dbase)
        self.assertTrue('MaxSplIntvl-RB' in dbase)
        self.assertTrue('SplIntvl-Mon' in dbase)
        self.assertTrue('SplIntvlBPM-Mon' in dbase)
        self.assertTrue('MinIntvlBtwSpl-SP' in dbase)
        self.assertTrue('MinIntvlBtwSpl-RB' in dbase)
        self.assertTrue('CurrOffset-SP' in dbase)
        self.assertTrue('CurrOffset-RB' in dbase)
        self.assertTrue('BuffRst-Cmd' in dbase)
        self.assertTrue('BuffAutoRst-Sel' in dbase)
        self.assertTrue('BuffAutoRst-Sts' in dbase)
        self.assertTrue('BuffAutoRstDCurr-SP' in dbase)
        self.assertTrue('BuffAutoRstDCurr-RB' in dbase)
        self.assertTrue('FrstSplTime-SP' in dbase)
        self.assertTrue('FrstSplTime-RB' in dbase)
        self.assertTrue('LastSplTime-SP' in dbase)
        self.assertTrue('LastSplTime-RB' in dbase)
        self.assertTrue('FrstSplTimeBPM-RB' in dbase)
        self.assertTrue('LastSplTimeBPM-RB' in dbase)
        self.assertTrue('Lifetime-Mon' in dbase)
        self.assertTrue('BuffSize-Mon' in dbase)
        self.assertTrue('BuffSizeTot-Mon' in dbase)
        self.assertTrue('BufferValue-Mon' in dbase)
        self.assertTrue('BufferTimestamp-Mon' in dbase)
        self.assertTrue('LifetimeBPM-Mon' in dbase)
        self.assertTrue('BuffSizeBPM-Mon' in dbase)
        self.assertTrue('BuffSizeTotBPM-Mon' in dbase)
        self.assertTrue('BufferValueBPM-Mon' in dbase)
        self.assertTrue('BufferTimestampBPM-Mon' in dbase)

        # PVs units
        self.assertEqual(dbase['MaxSplIntvl-SP']['unit'], 's')
        self.assertEqual(dbase['MaxSplIntvl-RB']['unit'], 's')
        self.assertEqual(dbase['SplIntvl-Mon']['unit'], 's')
        self.assertEqual(dbase['SplIntvlBPM-Mon']['unit'], 's')
        self.assertEqual(dbase['MinIntvlBtwSpl-SP']['unit'], 's')
        self.assertEqual(dbase['MinIntvlBtwSpl-RB']['unit'], 's')
        self.assertEqual(dbase['CurrOffset-SP']['unit'], 'mA')
        self.assertEqual(dbase['CurrOffset-RB']['unit'], 'mA')
        self.assertEqual(dbase['BuffAutoRstDCurr-SP']['unit'], 'mA')
        self.assertEqual(dbase['BuffAutoRstDCurr-RB']['unit'], 'mA')
        self.assertEqual(dbase['FrstSplTime-SP']['unit'], 's')
        self.assertEqual(dbase['FrstSplTime-RB']['unit'], 's')
        self.assertEqual(dbase['LastSplTime-SP']['unit'], 's')
        self.assertEqual(dbase['LastSplTime-RB']['unit'], 's')
        self.assertEqual(dbase['FrstSplTimeBPM-RB']['unit'], 's')
        self.assertEqual(dbase['LastSplTimeBPM-RB']['unit'], 's')
        self.assertEqual(dbase['Lifetime-Mon']['unit'], 's')
        self.assertEqual(dbase['LifetimeBPM-Mon']['unit'], 's')



"""."""
import time as _time
import logging as _logging

import numpy as _np
import matplotlib.pyplot as _mplt

from mathphys.functions import save as _save, load as _load, \
    get_namedtuple as _namedtuple
from .bpm_fam import FamBPMs as _FamBPMs
from .timing import Trigger
from .currinfo import CurrInfoSI as _CurrInfoSI


class EqualizeBPMs(_FamBPMs):
    """."""

    NR_POINTS = 2000
    MAX_MULTIPLIER = 0xffffff / (1 << 24)
    PROCMETHODSMEANING = {
        'AABS': 'All Antennas will be equalized in Both Semicycles',
        'EABS': 'Each Antenna will be equalized in Both Semicycles',
        'AAES': 'All Antennas will be equalized in Each Semicycle'
        }
    ProcMethods = _namedtuple('ProcMethods', ('AABS', 'EABS', 'AAES'))
    AcqStrategies = _namedtuple(
        'AcqStrategies', ('AssumeOrder', 'AcqInvRedGain'))

    def __init__(self, devname=None, bpmnames=None, logger=None):
        """."""
        self._logger = None
        if logger is not None:
            self.logger = logger
        self._proc_method = self.ProcMethods.EABS
        self._acq_strategy = self.AcqStrategies.AcqInvRedGain
        self._acq_inverse_reduced_gain = self.round_gains(0.95)
        self._acq_nrpoints = 2000
        self._acq_timeout = 30
        super().__init__(
            devname=devname, bpmnames=bpmnames, ispost_mortem=False,
            mturn_signals2acq='ABCD', props2init=[
                'SwDirGainA-RB', 'SwDirGainB-RB', 'SwDirGainC-RB',
                'SwDirGainD-RB', 'SwInvGainA-RB', 'SwInvGainB-RB',
                'SwInvGainC-RB', 'SwInvGainD-RB', 'SwDirGainA-SP',
                'SwInvGainA-SP', 'SwDirGainB-SP', 'SwInvGainB-SP',
                'SwDirGainC-SP', 'SwInvGainC-SP', 'SwDirGainD-SP',
                'SwInvGainD-SP', 'ACQTriggerEvent-Sel', 'ACQStatus-Sts',
                'GEN_AArrayData', 'GEN_BArrayData', 'GEN_CArrayData',
                'GEN_DArrayData', 'ACQTriggerRep-Sel', 'ACQChannel-Sel',
                'ACQTrigger-Sel', 'ACQSamplesPre-SP', 'ACQSamplesPost-SP',
                'INFOHarmonicNumber-RB', 'INFOTbTRate-RB', 'SwDivClk-RB',
                'ACQChannel-Sts', 'INFOFOFBRate-RB', 'PosKx-RB', 'PosKy-RB',
                'PosXOffset-RB', 'PosYOffset-RB'])
        self.currinfo = _CurrInfoSI(props2init=['Current-Mon', ])
        self.trigger = Trigger(
            'SI-Fam:TI-BPM', props2init=['Src-Sel', 'Src-Sts'])
        self._devices.append(self.currinfo)
        self._devices.append(self.trigger)
        self.data = dict()

    @property
    def logger(self):
        """Logger object. Must be an instance of `logging.Logger`."""
        return self._logger

    @logger.setter
    def logger(self, logger):
        if isinstance(logger, _logging.Logger):
            self._logger = logger
        else:
            self._log('ERR:Could no set logger. Wrong type.')

    @property
    def acq_strategy_str(self):
        """."""
        return self.AcqStrategies._fields[self._acq_strategy]

    @property
    def acq_strategy(self):
        """."""
        return self._acq_strategy

    @acq_strategy.setter
    def acq_strategy(self, val):
        val = self._enum_selector(val, self.AcqStrategies)
        if val is not None:
            self._acq_strategy = val

    @property
    def acq_inverse_reduced_gain(self):
        """."""
        return self._acq_inverse_reduced_gain

    @acq_inverse_reduced_gain.setter
    def acq_inverse_reduced_gain(self, value):
        self._acq_inverse_reduced_gain = self.round_gains(value)

    @property
    def acq_nrpoints(self):
        """."""
        return self._acq_nrpoints

    @acq_nrpoints.setter
    def acq_nrpoints(self, value):
        if value > 4:
            self._acq_nrpoints = int(value)

    @property
    def acq_timeout(self):
        """."""
        return self._acq_timeout

    @acq_timeout.setter
    def acq_timeout(self, value):
        if value > 0:
            self._acq_timeout = value

    @property
    def proc_method_str(self):
        """."""
        return self.ProcMethods._fields[self._proc_method]

    @property
    def proc_method(self):
        """."""
        return self._proc_method

    @proc_method.setter
    def proc_method(self, meth):
        meth = self._enum_selector(meth, self.ProcMethods)
        if meth is not None:
            self._proc_method = meth

    def load_data(self, filename):
        """Load self.data from file.

        Args:
            fname (str): name of the file. If extension is not provided,
                '.pickle' will be added and a pickle file will be assumed. If
                provided, must be {'.pickle', '.pkl'} for pickle files or
                {'.h5', '.hdf5', '.hdf', '.hd5'} for HDF5 files.

        """
        self.data = _load(filename, )

    def save_data(self, fname, overwrite=False, makedirs=True, compress=False):
        """Save self.data to pickle or HDF5 file.

        Args:
            fname (str): name of the file. If extension is not provided,
                '.pickle' will be added and a pickle file will be assumed. If
                provided, must be {'.pickle', .pkl} for pickle files or
                {'.h5', '.hdf5', '.hdf', '.hd5'} for HDF5 files.
            overwrite (bool, optional): Whether to overwrite existing file.
                Defaults to False.
            makedirs (bool, optional): create dir, if it does not exist.
                Defaults to False.
            compress (bool, optional): If True, the file will be saved in
                compressed format, using gzip library. Defaults to False.

        Raises:
            FileExistsError: in case overwrite is False and file exists.

        """
        _save(
            self.data, fname=fname, overwrite=overwrite, makedirs=makedirs,
            compress=compress)

    def get_current_gains(self):
        """Return Current BPM gains as 3D numpy array.

        Returns:
            numpy.ndarray (4, nrbpms, 2): 3D numpy array with gains.
                 - The first index vary the antennas: 'A', 'B', 'C', 'D';
                 - The second index vary the BPMs, in the default high level
                convention order;
                 - The last index refers to the type of the gain in the
                following order: 'Direct', 'Inverse';

        """
        gains = []
        for bpm in self.bpms:
            gaind = [getattr(bpm, f'gain_direct_{a}') for a in 'abcd']
            gaini = [getattr(bpm, f'gain_inverse_{a}') for a in 'abcd']
            gains.append([gaind, gaini])
        gains = _np.array(gains).swapaxes(0, 1).swapaxes(0, -1)
        return gains

    def set_gains(self, gains):
        """Set gains matrix to BPMs.

        Args:
            gains (float or numpy.ndarray, (4, nrbpms, 2)): gain matrix. In
                the same format as the one provided by get_current_gains
                method. If a float is provided, then this same gain will be
                applied to all BPMs.

        Raises:
            ValueError: If gains is a numpy array with wrong shape.

        """
        nbpm = len(self.bpms)
        shape = (4, nbpm, 2)
        if not isinstance(gains, _np.ndarray):
            gains = _np.full(shape, gains)
        if gains.shape != shape:
            raise ValueError(f'Wrong shape for gains. Must be {shape}')

        for j, ant in enumerate('abcd'):
            gns = gains[j]
            for i, bpm in enumerate(self.bpms):
                g = gns[i]
                setattr(bpm, f'gain_direct_{ant}', g[0])
                setattr(bpm, f'gain_inverse_{ant}', g[1])

    @staticmethod
    def round_gains(gains):
        """Make gains compatible with 24 bits fixed point representation."""
        mult = 1 << 24
        return _np.int_(gains*mult) / mult

    def acquire_bpm_data(self):
        """."""
        self._log('Starting Acquisition.')
        self.data = dict()
        self.data['timestamp'] = _time.time()
        self.data['stored_current'] = self.currinfo.current
        self.data['gains_init'] = self.get_current_gains()

        init_source = self.trigger.source
        ini_dly = self.trigger.delay_raw
        try:
            self._do_acquire()
        except Exception as err:
            self._log('ERR:Problem with acquisition:')
            self._log(f'ERR:{str(err)}')

        self.trigger.source = init_source
        self.trigger.delay_raw = ini_dly
        self.set_gains(self.data['gains_init'])
        self._log('Acquisition Finished!')

    def _do_acquire(self):
        if self._acq_strategy == self.AcqStrategies.AssumeOrder:
            self.set_gains(self.MAX_MULTIPLIER)
        else:
            gains = _np.full((4, len(self.bpms), 2), self.MAX_MULTIPLIER)
            gains[:, :, 1] *= self._acq_inverse_reduced_gain
            self.set_gains(gains)
        self.data['acq_strategy'] = self._acq_strategy
        self.data['acq_inverse_reduced_gain'] = self._acq_inverse_reduced_gain

        # acquire antennas data in FOFB rate
        self._log('Preparing BPMs')
        ret = self.cmd_mturn_acq_abort()
        if ret > 0:
            self._log(
                f'ERR: BPM {self.bpm_names[ret-1]} did not abort '
                'previous acquistion.')
            return

        self.mturn_reset_flags_and_update_initial_timestamps()
        ret = self.mturn_config_acquisition(
            nr_points_after=self._acq_nrpoints, nr_points_before=0,
            acq_rate='FOFB', repeat=False, external=True)
        if ret > 0:
            self._log(
                f'ERR: BPM {self.bpm_names[ret-1]} did not start acquistion.')
            return

        self.trigger.delay_raw = 0
        self.trigger.source = self.trigger.source_options.index('Clock3')

        self._log('Waiting BPMs to update')
        ret = self.mturn_wait_update(timeout=self._acq_timeout)
        if ret > 0:
            self._log(
                f'ERR: BPM {self.bpm_names[ret-1]} did not update in time.')
            return
        elif ret < 0:
            self._log(f'ERR: Problem with acquisition. Error code {ret}')
            return
        self._log('BPMs updated.')

        self._log('Acquiring data.')
        fswtc = self.get_switching_frequency(1)
        fsamp = self.get_sampling_frequency(1)
        self.data['freq_switching'] = fswtc
        self.data['freq_sampling'] = fsamp
        if None in {fsamp, fswtc}:
            self._log('ERR: Not all BPMs are configured equally.')
            return
        elif fsamp % (2*fswtc):
            self._log('ERR: Sampling freq is not multiple of switching freq.')
            return

        posx_gain, posy_gain = [], []
        posx_offset, posy_offset = [], []
        for bpm in self.bpms:
            posx_gain.append(bpm.posx_gain)
            posy_gain.append(bpm.posy_gain)
            posx_offset.append(bpm.posx_offset)
            posy_offset.append(bpm.posy_offset)
        self.data['posx_gain'] = _np.array(posx_gain)
        self.data['posy_gain'] = _np.array(posy_gain)
        self.data['posx_offset'] = _np.array(posx_offset)
        self.data['posy_offset'] = _np.array(posy_offset)
        self.data['gains_acq'] = self.get_current_gains()

        _time.sleep(0.1)
        self.data['antennas'] = _np.array(
            self.get_mturn_signals()).swapaxes(-1, -2)

    def acquire_data_for_checking(self):
        """."""
        self._log('Starting Acquisition.')

        init_source = self.trigger.source
        ini_dly = self.trigger.delay_raw
        try:
            self._do_acquire_for_check()
        except Exception as err:
            self._log('ERR:Problem with acquisition:')
            self._log(f'ERR:{str(err)}')

        self.trigger.source = init_source
        self.trigger.delay_raw = ini_dly
        self._log('Acquisition Finished!')

    def _do_acquire_for_check(self):
        # acquire antennas data in FOFB rate
        self._log('Preparing BPMs')
        ret = self.cmd_mturn_acq_abort()
        if ret > 0:
            self._log(
                f'ERR: BPM {self.bpm_names[ret-1]} did not abort '
                'previous acquistion.')
            return

        fswtc = self.get_switching_frequency(1)
        fsamp = self.get_sampling_frequency(1)
        nrpts = int(fsamp / fswtc)

        self.mturn_reset_flags_and_update_initial_timestamps()
        ret = self.mturn_config_acquisition(
            nr_points_after=nrpts, nr_points_before=0,
            acq_rate='FOFB', repeat=False, external=True)
        if ret > 0:
            self._log(
                f'ERR: BPM {self.bpm_names[ret-1]} did not start acquistion.')
            return

        self.trigger.delay_raw = 0
        self.trigger.source = self.trigger.source_options.index('Clock3')

        self._log('Waiting BPMs to update')
        ret = self.mturn_wait_update(timeout=self._acq_timeout)
        if ret > 0:
            self._log(
                f'ERR: BPM {self.bpm_names[ret-1]} did not update in time.')
            return
        elif ret < 0:
            self._log(f'ERR: Problem with acquisition. Error code {ret}')
            return
        self._log('BPMs updated.')

        self._log('Acquiring data.')
        if None in {fsamp, fswtc}:
            self._log('ERR: Not all BPMs are configured equally.')
            return
        elif fsamp % (2*fswtc):
            self._log('ERR: Sampling freq is not multiple of switching freq.')
            return

        _time.sleep(0.1)
        self.data['antennas_for_check'] = _np.array(
            self.get_mturn_signals()).swapaxes(-1, -2)

    # --------- Methods for processing ------------

    def process_data(self):
        """Process data."""
        if 'antennas' not in self.data:
            self._log('ERR:There is no data to process. Acquire first.')
            return
        mean_antennas = self.calc_switching_levels(**self.data)
        self.calc_gains(mean_antennas)
        self.estimate_orbit_variation()

    def calc_switching_levels(
            self, antennas, fsamp=4, fswtc=1, acq_strategy=0,
            acq_inverse_reduced_gain=0.95, **kwargs):
        """Calculate average signal for each antenna in both switching states.

        Args:
            antennas (numpy.ndarray, (4, nrbpms, N)): Antennas data.
            fsamp (float, optional): Sampling frequency. Defaults to 4.
            fswtc (float, optional): Switching frequency. Defaults to 1.
            acq_strategy (int, optional): Whether we should assume states
                order in data, starting with the direct state, or if we should
                look for the higher levels (direct) and lower levels
                (inverse). Defaults to 0.
            acq_inverse_reduced_gain (float, optional): Level of the inverse
                state, if acq_strategy is 1. Must be the same value used in
                antennas gain during acquisition. Defaults to 0.95.

        Returns:
            mean (numpy.ndarray, (4, nrbpms, 2)): The levels of each switching
                state for each antenna.

        """
        _ = kwargs
        ants = antennas
        lsemicyc = int(fsamp // (2*fswtc))
        nbpm = len(self.bpms)
        nant = 4
        trunc = (ants.shape[-1] // (lsemicyc*2)) * (lsemicyc*2)
        if trunc < 5:
            self._log(
                f'ERR:Data not large enough. Acquire data with more points.')
            return
        elif ants.shape[-1] != trunc:
            ants = ants[:, :, :trunc]
            self._log(f'WARN:Truncating data at {trunc} points')
        ants = ants.reshape(nant, nbpm, -1, lsemicyc*2)
        ants = ants.mean(axis=2)
        self._log('Calculating switching levels.')
        self._log(
            f'AcqStrategy is {self.AcqStrategies._fields[acq_strategy]}.')
        if acq_strategy == self.AcqStrategies.AssumeOrder:
            ants = ants.reshape(nant, nbpm, 2, lsemicyc)
            mean = ants.mean(axis=-1)
            # Inverse comes first!:
            mean = mean[:, :, ::-1]
            idn = _np.tile(_np.arange(lsemicyc), (nbpm, 1))
            idp = idn + lsemicyc
        else:
            # Try to find out the two states by looking at different levels in
            # the sum of the four antennas of each BPM.
            dts = ants.sum(axis=0)
            idcs = dts - dts.mean(axis=-1)[..., None] > 0
            idp = idcs.nonzero()  # direct
            idn = (~idcs).nonzero()  # inverse
            cond = idp[0].size == nbpm*lsemicyc
            cond &= _np.unique(idp[0]).size == nbpm
            if not cond:
                self._log(
                    'ERR: Could not identify switching states appropriately.')
                return
            mean = _np.zeros((nant, nbpm, 2))
            dtp = ants[:, idp[0], idp[1]].reshape(nant, nbpm, lsemicyc)
            dtn = ants[:, idn[0], idn[1]].reshape(nant, nbpm, lsemicyc)
            mean[:, :, 0] = dtp.mean(axis=-1)
            mean[:, :, 1] = dtn.mean(axis=-1)
            # Re-scale the inverse state data to match the direct state:
            scl = self.MAX_MULTIPLIER / acq_inverse_reduced_gain
            mean[:, :, 1] *= scl
            idp = idp[1].reshape(nbpm, -1)
            idn = idn[1].reshape(nbpm, -1)
        self.data['antennas_mean'] = mean
        self.data['idcs_direct'] = idp
        self.data['idcs_inverse'] = idn
        return mean

    def calc_gains(self, mean):
        """Calculate gains given mean data from antennas for both semi-cycles.

        Args:
            mean (numpy.array, (nrbpms, 4, 2)): Mean data from antennas for
                both switching semi-cycle levels. Compatible with output of
                calc_switching_levels.

        """
        maxm = self.MAX_MULTIPLIER
        self._log('Calculating Gains.')
        self._log(
            f'ProcMethod is {self.ProcMethods._fields[self._proc_method]}')
        if self._proc_method == self.ProcMethods.EABS:
            # equalize each antenna for both semicycles
            min_ant = mean.min(axis=-1)
            min_ant = min_ant[:, :, None]
        elif self._proc_method == self.ProcMethods.AABS:
            # equalize the 4 antennas for both semicycles
            min_ant = mean.min(axis=0).min(axis=-1)
            min_ant = min_ant[None, :, None]
        elif self._proc_method == self.ProcMethods.AAES:
            # equalize the 4 antennas for each semicycle
            min_ant = mean.min(axis=0)
            min_ant = min_ant[None, :, :]
        min_ant *= maxm
        gains = self.round_gains(min_ant / mean)
        self.data['proc_method'] = self._proc_method
        self.data['gains_new'] = gains
        return gains

    def estimate_orbit_variation(self):
        """Estimate orbit variation between old and new gains."""
        self._log('Estimating Orbit Variation.')
        mean = self.data.get('antennas_mean')
        gains_init = self.data.get('gains_init')
        gains_new = self.data.get('gains_new')
        gainx = self.data.get('posx_gain')
        gainy = self.data.get('posy_gain')
        offx = self.data.get('posx_offset', _np.zeros(len(self.bpms)))
        offy = self.data.get('posy_offset', _np.zeros(len(self.bpms)))
        if gains_new is None:
            self._log('ERR:Missing info. Acquire and process data first.')

        orbx_init, orby_init = self._estimate_orbit(
            mean, gains_init, gainx, gainy, offx, offy)
        orbx_new, orby_new = self._estimate_orbit(
            mean, gains_new, gainx, gainy, offx, offy)
        # Get the average over both semicycles
        self.data['orbx_init'] = orbx_init.mean(axis=-1)
        self.data['orby_init'] = orby_init.mean(axis=-1)
        self.data['orbx_new'] = orbx_new.mean(axis=-1)
        self.data['orby_new'] = orby_new.mean(axis=-1)
        dorbx = orbx_new - orbx_init
        dorby = orby_new - orby_init
        self.data['dorbx'] = dorbx.mean(axis=-1)
        self.data['dorby'] = dorby.mean(axis=-1)
        return dorbx, dorby

    # ---------------------- plot methods -------------------------------------

    def plot_orbit_distortion(self):
        """."""
        dorbx = self.data.get('dorbx')
        dorby = self.data.get('dorby')
        if dorbx is None:
            self._log('ERR:Must acquire data and process first.')
            return None, None

        fig, ax = _mplt.subplots(1, 1, figsize=(5, 3))
        ax.plot(dorbx, label=r'Horizontal')
        ax.plot(dorby, label=r'Vertical')
        ax.set_ylabel(r'$\Delta$ Orbit [$\mu$m]')
        ax.set_xlabel('BPM Index')
        ax.grid(True, alpha=0.5, ls='--', lw=1)
        ax.set_title('Orbit Variation: New - Old')
        ax.legend(loc='best', fontsize='x-small', ncol=2)
        fig.tight_layout()
        return fig, ax

    def plot_gains(self):
        """."""
        gini = self.data.get('gains_init')
        gnew = self.data.get('gains_new')
        if gnew is None:
            self._log('ERR:Must acquire data and process first.')
            return None, None

        fig, axs = _mplt.subplots(
            4, 2, figsize=(9, 8), sharex=True, sharey=True)
        ants = 'ABCD'
        for i in range(4):
            ldo = axs[i, 0].plot(gini[i, :, 0])[0]
            lio = axs[i, 1].plot(gini[i, :, 1], color=ldo.get_color())[0]
            ldn = axs[i, 0].plot(gnew[i, :, 0])[0]
            lin = axs[i, 1].plot(gnew[i, :, 1], color=ldn.get_color())[0]
            if not i:
                ldo.set_label('Old')
                ldn.set_label('New')
            axs[i, 0].set_ylabel(ants[i])
            axs[i, 0].grid(True, alpha=0.5, ls='--', lw=1)
            axs[i, 1].grid(True, alpha=0.5, ls='--', lw=1)

        axs[0, 0].set_title('Direct Gains')
        axs[0, 1].set_title('Inverse Gains')
        axs[0, 0].legend(loc='best')
        axs[-1, 0].set_xlabel('BPM Index')
        axs[-1, 1].set_xlabel('BPM Index')
        fig.tight_layout()
        return fig, axs

    def plot_semicycle_indices(self):
        """."""
        idp = self.data.get('idcs_direct')
        idn = self.data.get('idcs_inverse')
        acqs = self.data.get('acq_strategy')
        if idp is None:
            self._log('ERR:Must acquire data and process first.')
            return None, None

        fig, ax = _mplt.subplots(1, 1, figsize=(5, 3))
        ax.plot(idp, 'o', color='C0')[0].set_label('Direct')
        ax.plot(idn, 'o', color='C1')[0].set_label('Inverse')
        ax.set_ylabel('Switching Cycle Indices')
        ax.set_xlabel('BPM Index')
        ax.grid(axis='x', alpha=0.5, ls='--', lw=1)
        ax.legend(loc='best', ncol=2, fontsize='x-small')
        ax.set_title(f'Acq Strategy: ' + self.AcqStrategies._fields[acqs])
        fig.tight_layout()
        return fig, ax

    def plot_antennas_mean(self):
        """."""
        gainx = self.data.get('posx_gain')
        gainy = self.data.get('posx_gain')
        offx = self.data.get('posx_offset')
        offy = self.data.get('posx_offset')
        gacq = self.data.get('gains_acq')
        gini = self.data.get('gains_init')
        gnew = self.data.get('gains_new')
        antm = self.data.get('antennas_mean')
        if gnew is None:
            self._log('ERR:Must acquire data and process first.')
            return None, None

        gains = (1, gacq, gini, gnew)
        titles = ('Unit Gain', 'Acquisition Gain', 'Old Gain', 'New Gain')
        labs = list('ABCDXYS')
        labs[-1] = 'Sum'
        labs = [lab + ' [a.u.]' for lab in labs]
        labs[4] = r'X [$\mu$m]'
        labs[5] = r'Y [$\mu$m]'

        antm /= 10**(_np.floor(_np.log10(antm.max())))

        fig, axs = _mplt.subplots(
            len(labs), len(gains), figsize=(16, 10), sharex=True, sharey='row')
        for j, gain in enumerate(gains):
            val = antm * gain
            for i in range(4):
                ax = axs[i, j]
                ld = ax.plot(val[i, :, 0], 'o-')[0]
                li = ax.plot(val[i, :, 1], 'o-')[0]
                if not i and not j:
                    ld.set_label('Direct')
                    li.set_label('Inverse')
            posx, posy = self._estimate_orbit(
                antm, gain, gainx, gainy, offx, offy)
            sum_ = val.sum(axis=0)
            axs[4, j].plot(posx[:, 0], 'o-')
            axs[4, j].plot(posx[:, 1], 'o-')
            axs[5, j].plot(posy[:, 0], 'o-')
            axs[5, j].plot(posy[:, 1], 'o-')
            axs[6, j].plot(sum_[:, 0], 'o-')
            axs[6, j].plot(sum_[:, 1], 'o-')
            for i, lab in enumerate(labs):
                ax = axs[i, j]
                if not j:
                    ax.set_ylabel(lab)
                ax.grid(True, alpha=0.5, ls='--', lw=1)
            axs[0, j].set_title(titles[j])
            axs[-1, j].set_xlabel('BPM Index')

        axs[0, 0].legend(loc='best', fontsize='x-small', ncol=2)
        fig.tight_layout()
        return fig, axs

    def plot_antennas_for_check(self):
        """."""
        antd = self.data.get('antennas_for_check')
        if antd is None:
            self._log('ERR:Must acquire data for check first.')
            return None, None

        fig, axs = _mplt.subplots(
            4, 1, figsize=(6, 8), sharex=True, sharey=True)
        ants = 'ABCD'
        mean = antd.mean(axis=-1)[:, :, None]
        antd = antd / mean - 1
        nbpm = antd.shape[1]
        for j in range(nbpm):
            cor = _mplt.cm.jet(j/(nbpm-1))
            for i in range(4):
                axs[i].plot(antd[i, j], 'o-', color=cor)
                if not j:
                    axs[i].set_ylabel(ants[i] + ' [%]')
                    axs[i].grid(True, alpha=0.5, ls='--', lw=1)

        axs[0].set_title('Relative Variation')
        axs[-1].set_xlabel('Samples')
        fig.tight_layout()
        return fig, axs

    # ------- auxiliary methods ----------

    @staticmethod
    def _estimate_orbit(mean, gains, gainx, gainy, offx, offy):
        a, b, c, d = mean * gains
        # Calculate difference over sum for each pair
        a_c = (a-c) / (a+c)
        b_d = (b-d) / (b+d)
        # Get the positions:
        posx = (a_c - b_d) / 2
        posy = (a_c + b_d) / 2 
        # Apply position gains:
        posx *= gainx[:, None]
        posy *= gainy[:, None]
        # Subtract offsets:
        posx -= offx[:, None]
        posy -= offy[:, None]
        return posx, posy

    def _log(self, message, *args, level='INFO', **kwrgs):
        if self._logger is None:
            print(message, *args, **kwrgs)
        elif message.startswith('WARN:'):
            self._logger.warning(message[5:])
        elif message.startswith('ERR:'):
            self._logger.error(message[4:])
        else:
            self._logger.info(message)



"""Epics Devices and Device Application."""

import time as _time
import operator as _opr
import math as _math
from functools import partial as _partial

from epics.ca import ChannelAccessGetFailure as _ChannelAccessGetFailure, \
    CASeverityException as _CASeverityException
import numpy as _np

from ..envars import VACA_PREFIX as _VACA_PREFIX
from ..epics import PV as _PV, CONNECTION_TIMEOUT as _CONN_TIMEOUT, \
    GET_TIMEOUT as _GET_TIMEOUT
from ..simul import SimPV as _PVSim
from ..simul import Simulation as _Simulation
from ..namesys import SiriusPVName as _SiriusPVName

_DEF_TIMEOUT = 10  # s
_TINY_INTERVAL = 0.050  # s


class Device:
    """Epics Device.

    Parameters
    ----------
        devname: str
            Device name, to be used as PVs prefix.
        props2init: ('all', None or (tuple, list)), optional
            Define which PVs will be connected in the class instantiation. If
            equal to 'all', all properties listed in PROPERTIES_DEFAULT will
            be initialized. If None or empty iterable, no property will be
            initialized. If iterable of strings, only the properties listed
            will be initialized. In this last option, it is possible to add
            properties that are not in PROPERTIES_DEFAULT. The other properties
            will be created whenever they are needed. Defaults to 'all'.
        auto_monitor: bool, optional
            Whether to automatically monitor PVs for changes. Used for PVs
            that do not end with '-Mon' or 'Data'. Defaults to True.
        auto_monitor_mon: bool, optional
            Whether to automatically monitor '-Mon' or 'Data' PVs for changes.
            Defaults to False (to avoid overloading the client). Set to False
            when using PV.get_timevars() to know when a PV has been updated.

    """

    CONNECTION_TIMEOUT = _CONN_TIMEOUT
    GET_TIMEOUT = _GET_TIMEOUT
    PROPERTIES_DEFAULT = ()

    PROPERTY_SEP = ':'

    def __init__(
            self, devname, props2init='all', auto_monitor=True,
            auto_monitor_mon=False):
        """."""
        self._devname = _SiriusPVName(devname) if devname else devname
        self._auto_monitor = auto_monitor
        self._auto_monitor_mon = auto_monitor_mon

        if isinstance(props2init, str) and props2init.lower() == 'all':
            propties = self.PROPERTIES_DEFAULT
        elif not props2init:
            propties = []
        elif isinstance(props2init, (list, tuple)):
            propties = props2init
        else:
            raise ValueError('Wrong value for props2init.')
        self._pvs = {prpt: self._create_pv(prpt) for prpt in propties}

    @property
    def devname(self):
        """Return device name."""
        return self._devname

    @property
    def properties_in_use(self):
        """Return properties that were already added to the PV list."""
        return tuple(sorted(self._pvs.keys()))

    @property
    def properties_added(self):
        """Return properties that were added to the PV list that are not in
        PROPERTIES_DEFAULT."""
        return tuple(sorted(
            set(self.properties_in_use) - set(self.PROPERTIES_DEFAULT)))

    @property
    def properties_all(self):
        """Return all properties of the device, connected or not."""
        return tuple(sorted(
            set(self.PROPERTIES_DEFAULT + self.properties_in_use)))

    @property
    def simulators(self):
        """Return simulator."""
        sims = set()
        for pvname in self.pvnames:
            sims.update(_Simulation.simulator_find(pvname))
        return sims

    @property
    def pvnames(self):
        """Return device PV names."""
        pvnames = {pv.pvname for pv in self._pvs.values()}
        return pvnames

    @property
    def connected(self):
        """Return PVs connection status."""
        for pvobj in self._pvs.values():
            if not pvobj.connected:
                return False
        return True

    @property
    def auto_monitor_status(self):
        """Return PVs auto_monitor statuses."""
        return {pvn: pv.auto_monitor for pvn, pv in self._pvs.items()}

    @property
    def disconnected_pvnames(self):
        """Return list of disconnected device PVs."""
        set_ = set()
        for pvobj in self._pvs.values():
            if not pvobj.connected:
                set_.add(pvobj.pvname)
        return set_

    def set_auto_monitor(self, pvname, value):
        """Set auto_monitor state of individual PVs."""
        if pvname not in self._pvs:
            return False
        pvobj = self._pvs[pvname]
        try:
            # TODO verify need of int
            pvobj.auto_monitor = int(value)
        except (_ChannelAccessGetFailure, _CASeverityException):
            # exceptions raised in a Virtual Circuit Disconnect (192)
            # event. If the PV IOC goes down, for example.
            print('Could not set auto_monitor of {}'.format(pvobj.pvname))
            return False
        return True

    def update(self):
        """Update device properties."""
        for pvobj in self._pvs.values():
            pvobj.get()

    def pv_object(self, propty):
        """Return PV object for a given device property."""
        if propty not in self._pvs:
            pvobj = self._create_pv(propty)
            pvobj.wait_for_connection(Device.CONNECTION_TIMEOUT)
            self._pvs[propty] = pvobj
        return self._pvs[propty]

    def pv_ctrlvars(self, propty):
        """Return PV object control variable."""
        return self._pvs[propty].get_ctrlvars()

    def pv_attribute_values(self, attribute):
        """Return pvname-value dict of a given attribute for all PVs."""
        attributes = dict()
        for pvobj in self._pvs.values():
            attributes[pvobj.pvname] = getattr(pvobj, attribute)
        return attributes

    @property
    def hosts(self):
        """Return dict of IOC hosts providing device properties."""
        return self.pv_attribute_values('host')

    @property
    def values(self):
        """Return dict of property values."""
        return self.pv_attribute_values('value')

    def wait_for_connection(self, timeout=None):
        """Wait for connection."""
        for pvobj in self._pvs.values():
            res = pvobj.wait_for_connection(timeout)
            if not res:
                return False
        return True

    def __getitem__(self, propty):
        """Return value of property."""
        pvobj = self.pv_object(propty)
        try:
            value = pvobj.get(timeout=Device.GET_TIMEOUT)
        except (_ChannelAccessGetFailure, _CASeverityException):
            # exceptions raised in a Virtual Circuit Disconnect (192)
            # event. If the PV IOC goes down, for example.
            print('Could not get value of {}'.format(pvobj.pvname))
            value = None
        return value

    def __setitem__(self, propty, value):
        """Set value of property."""
        pvobj = self.pv_object(propty)
        try:
            pvobj.value = value
        except (_ChannelAccessGetFailure, _CASeverityException):
            print('Could not set value of {}'.format(pvobj.pvname))

    # --- private methods ---
    def _create_pv(self, propty):
        pvname = self._get_pvname(propty)
        auto_monitor = self._auto_monitor
        if pvname.endswith(('-Mon', 'Data')):
            auto_monitor = self._auto_monitor_mon
        in_sim = _Simulation.pv_check(pvname)
        pvclass = _PVSim if in_sim else _PV
        return pvclass(
            pvname, auto_monitor=auto_monitor,
            connection_timeout=Device.CONNECTION_TIMEOUT)

    def _wait(self, propty, value, timeout=None, comp='eq'):
        """."""
        def comp_(val):
            boo = comp(self[propty], val)
            if isinstance(boo, _np.ndarray):
                boo = _np.all(boo)
            return boo

        if isinstance(comp, str):
            comp = getattr(_opr, comp)

        if not isinstance(timeout, str) and timeout != 'never':
            timeout = _DEF_TIMEOUT if timeout is None else timeout
            timeout = 0 if timeout <= 0 else timeout
        t0_ = _time.time()
        while not comp_(value):
            if isinstance(timeout, str) and timeout == 'never':
                pass
            else:
                if _time.time() - t0_ > timeout:
                    return False
            _time.sleep(_TINY_INTERVAL)
        return True

    def _wait_float(
            self, propty, value, rel_tol=0.0, abs_tol=0.1, timeout=None):
        """Wait until float value gets close enough of desired value."""
        isc = _np.isclose if isinstance(value, _np.ndarray) else _math.isclose
        func = _partial(isc, abs_tol=abs_tol, rel_tol=rel_tol)
        return self._wait(propty, value, comp=func, timeout=timeout)

    def _wait_set(self, props_values, timeout=None, comp='eq'):
        timeout = _DEF_TIMEOUT if timeout is None else timeout
        t0_ = _time.time()
        for propty, value in props_values.items():
            timeout_left = max(0, timeout - (_time.time() - t0_))
            if timeout_left == 0:
                return False
            if not self._wait(propty, value, timeout=timeout_left, comp=comp):
                return False
        return True

    def _get_pvname(self, propty):
        dev = self._devname
        pref = _VACA_PREFIX + ('-' if _VACA_PREFIX else '')
        if isinstance(dev, _SiriusPVName) and \
                dev.is_standard(name_type='devname'):
            ppt = dev.propty
            pvname = dev.substitute(prefix=_VACA_PREFIX, propty=ppt + propty)
        elif dev:
            pvname = pref + dev + self.PROPERTY_SEP + propty
        else:
            pvname = pref + propty
        return pvname

    def _enum_setter(self, propty, value, enums):
        value = self._enum_selector(value, enums)
        if value is not None:
            self[propty] = value

    @staticmethod
    def _enum_selector(value, enums):
        if value is None:
            return
        if hasattr(enums, '_fields'):
            enums = enums._fields
        if isinstance(value, str) and value in enums:
            return enums.index(value)
        elif 0 <= int(value) < len(enums):
            return value


class DeviceSet:
    """."""

    def __init__(self, devices, devname=''):
        """."""
        self._devices = devices
        self._devname = _SiriusPVName(devname)

    _enum_selector = staticmethod(Device._enum_selector)

    @property
    def devname(self):
        """Name of the Device set. May be empty in some cases."""
        return self._devname

    @property
    def simulators(self):
        """Return list of simulators."""
        sims = set()
        for dev in self._devices:
            sims.update(dev.simulators)
        return sims

    @property
    def pvnames(self):
        """Return device PV names."""
        set_ = set()
        for dev in self._devices:
            set_.update(dev.pvnames)
        return set_

    @property
    def connected(self):
        """Return PVs connection status."""
        for dev in self._devices:
            if not dev.connected:
                return False
        return True

    @property
    def disconnected_pvnames(self):
        """Return list of disconnected device PVs."""
        set_ = set()
        for dev in self._devices:
            set_.update(dev.disconnected_pvnames)
        return set_

    @property
    def auto_monitor_status(self):
        """Return PVs auto_monitor statuses."""
        dic_ = dict()
        for dev in self._devices:
            dic_.update(dev.auto_monitor_status)
        return dic_

    def update(self):
        """Update device properties."""
        for dev in self._devices:
            dev.update()

    def pv_attribute_values(self, attribute):
        """Return property-value dict of a given attribute for all PVs."""
        attributes = dict()
        for dev in self._devices:
            attrs = dev.pv_attribute_values(attribute)
            attributes.update(attrs)
        return attributes

    @property
    def hosts(self):
        """Return dict of IOC hosts providing device properties."""
        return self.pv_attribute_values('host')

    @property
    def values(self):
        """Return dict of property values."""
        return self.pv_attribute_values('value')

    def wait_for_connection(self, timeout=None):
        """Wait for connection."""
        for dev in self._devices:
            if not dev.wait_for_connection(timeout=timeout):
                return False
        return True

    @property
    def devices(self):
        """Return devices."""
        return self._devices

    # --- private methods ---

    def _set_devices_propty(self, devices, propty, values, wait=0):
        """Set devices property to value(s)."""
        dev2val = self._get_dev_2_val(devices, values)
        for dev, val in dev2val.items():
            if dev.pv_object(propty).wait_for_connection():
                dev[propty] = val
                _time.sleep(wait)

    def _wait_devices_propty(
            self, devices, propty, values, comp='eq',
            timeout=None, return_prob=False):
        """Wait for devices property to reach value(s)."""
        if isinstance(comp, str):
            comp = getattr(_opr, comp)
        dev2val = self._get_dev_2_val(devices, values)

        timeout = _DEF_TIMEOUT if timeout is None else timeout
        tini = _time.time()
        for _ in range(int(timeout/_TINY_INTERVAL)):
            okdevs = set()
            for k, v in dev2val.items():
                boo = comp(k[propty], v)
                if isinstance(boo, _np.ndarray):
                    boo = _np.all(boo)
                if boo:
                    okdevs.add(k)
                if _time.time() - tini > timeout:
                    break
            list(map(dev2val.__delitem__, okdevs))
            if not dev2val:
                break
            if _time.time() - tini > timeout:
                break
            _time.sleep(_TINY_INTERVAL)

        allok = not dev2val
        if return_prob:
            return allok, [dev.pv_object(propty).pvname for dev in dev2val]
        return allok

    def _get_dev_2_val(self, devices, values):
        """Get devices to values dict."""
        # always use an iterable object
        if not isinstance(devices, (tuple, list)):
            devices = [devices, ]
        # if 'values' is not iterable, consider the same value for all devices
        if not isinstance(values, (tuple, list)):
            values = len(devices)*[values]
        return {k: v for k, v in zip(devices, values)}

    def __getitem__(self, devidx):
        """Return device."""
        return self._devices[devidx]



"""SI Insertion Device Feedforward (IDFF) Configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# IDFF corrector table
#
# 1st-level keys:
#   - 'pvnames': its value is a dict with
#       ('pparameter', 'kparameter', 'ch1', 'ch2', ... 'qs1', 'qs2') as
#       keys and corresponding pvname values for
#       ID k-variation parameter readout and correctors setpoints.
#   - 'polarizations': its value is a dict with
#       ('none', 'horizontal', 'vertical', ...) light polarization configs of
#       the ID as keys and dicts with kparameter and correctors
#       IDFF tables as lists.
#
#   obs:
#   1) pparameter: pvname that varies polarization configurations
#   2) kparameter: pvname that varies k (photon energy) configurations
#   3) 'none' polarization configuration should always be in the config.
#       it is a transition configuration between polarization configs.

_template_dict = {
    'description': 'description of the IDFF particularities.',
    'pvnames': {
        'pparameter': 'SI-10SB:ID-EPU50:Phase-Mon',
        'kparameter': 'SI-10SB:ID-EPU50:Gap-Mon',
        'ch1': 'SI-10SB:PS-CH-1:Current-SP',
        'ch2': 'SI-10SB:PS-CH-2:Current-SP',
        'cv1': 'SI-10SB:PS-CV-1:Current-SP',
        'cv2': 'SI-10SB:PS-CV-2:Current-SP',
        'qs1': 'SI-10SB:PS-QS-1:Current-SP',
        'qs2': 'SI-10SB:PS-QS-2:Current-SP',
    },
    'polarizations': {
        'none': {
            'pparameter': [-16.39, 0.0, 16.39, 25.0],
            'kparameter': 300,
            'ch1': 4*[0],
            'ch2': 4*[0],
            'cv1': 4*[0],
            'cv2': 4*[0],
            'qs1': 4*[0],
            'qs2': 4*[0],
            },
        'circularn': {
            'pparameter': -16.39,
            'kparameter': [22.0, 300.0],
            'ch1': 2*[0],
            'ch2': 2*[0],
            'cv1': 2*[0],
            'cv2': 2*[0],
            'qs1': 2*[0],
            'qs2': 2*[0],
            },
        'horizontal': {
            'pparameter': 0,
            'kparameter': [22.0, 100.0, 300.0],
            'ch1': 3*[0],
            'ch2': 3*[0],
            'cv1': 3*[0],
            'cv2': 3*[0],
            'qs1': 3*[0],
            'qs2': 3*[0],
            },
        'circularp': {
            'pparameter': 16.39,
            'kparameter': [22.0, 300.0],
            'ch1': 2*[0],
            'ch2': 2*[0],
            'cv1': 2*[0],
            'cv2': 2*[0],
            'qs1': 2*[0],
            'qs2': 2*[0],
            },
        'vertical': {
            'pparameter': 25,
            'kparameter': [22.0, 60, 100.0, 200, 300.0],
            'ch1': 5*[0],
            'ch2': 5*[0],
            'cv1': 5*[0],
            'cv2': 5*[0],
            'qs1': 5*[0],
            'qs2': 5*[0],
            },
    },
}



#!/usr/local/bin/python-sirius
"""Driver module."""

from pcaspy import Alarm as _Alarm, Severity as _Severity

from ...namesys import SiriusPVName
from ..app import App as _App
from ..pvs import ComputedPV as _ComputedPV

from .csdev import Const as _Const
from .pvs import BORFStatusPV as _BORFStatusPV, \
    SIRFStatusPV as _SIRFStatusPV, \
    SICheckAmpErrPV as _SICheckAmpErrPV, \
    SICheckPhsErrPV as _SICheckPhsErrPV, \
    SICheckDTuneErrPV as _SICheckDTuneErrPV


class RFDiagApp(_App):
    """Main application responsible for updating DB."""

    def _create_computed_pvs(self, *args):
        pref = self._prefix + ('-' if self._prefix else '')

        # BO
        devname = SiriusPVName(_Const.BO_DEV)

        # DiagStatus-Mon
        pvs = [None]*4
        pvs[_BORFStatusPV.PV_SIRIUS_INTLK] = \
            pref+'RA-RaBO02:RF-IntlkCtrl:IntlkSirius-Mon'
        pvs[_BORFStatusPV.PV_LLRF_INTLK] = pref+'RA-RaBO01:RF-LLRF:Intlk-Mon'
        pvs[_BORFStatusPV.PV_RMP_ENBLD] = pref+'BR-RF-DLLRF-01:RmpEnbl-Sts'
        pvs[_BORFStatusPV.PV_RMP_READY] = pref+'BR-RF-DLLRF-01:RmpReady-Mon'
        pvo = _ComputedPV(
            devname + ':DiagStatus-Mon', _BORFStatusPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        # SI
        devname = SiriusPVName(_Const.SI_DEV)
        prefname = devname.substitute(prefix=self._prefix)

        # DiagAmpErrSts-Mon
        pvs = [None]*1
        pvs[_SICheckAmpErrPV.PV_ERR] = pref+'SR-RF-DLLRF-01:SL:ERR:AMP'
        pvo = _ComputedPV(
            devname + ':DiagAmpErrSts-Mon', _SICheckAmpErrPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        # DiagPhsErrSts-Mon
        pvs = [None]*1
        pvs[_SICheckPhsErrPV.PV_ERR] = pref+'SR-RF-DLLRF-01:SL:ERR:PHS'
        pvo = _ComputedPV(
            devname + ':DiagPhsErrSts-Mon', _SICheckPhsErrPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        # DiagDTuneErrSts-Mon
        pvs = [None]*1
        pvs[_SICheckDTuneErrPV.PV_ERR] = pref+'SR-RF-DLLRF-01:TUNE:DEPHS'
        pvo = _ComputedPV(
            devname + ':DiagDTuneErrSts-Mon', _SICheckDTuneErrPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        # DiagStatus-Mon
        pvs = [None]*5
        pvs[_SIRFStatusPV.PV_SIRIUS_INTLK] = \
            pref+'RA-RaSIA02:RF-IntlkCtrl:IntlkSirius-Mon'
        pvs[_SIRFStatusPV.PV_LLRF_INTLK] = pref+'RA-RaSIA01:RF-LLRF:Intlk-Mon'
        pvs[_SIRFStatusPV.PV_AMPL_ERR] = prefname + ':DiagAmpErrSts-Mon'
        pvs[_SIRFStatusPV.PV_PHSE_ERR] = prefname + ':DiagPhsErrSts-Mon'
        pvs[_SIRFStatusPV.PV_DTUN_ERR] = prefname + ':DiagDTuneErrSts-Mon'
        pvo = _ComputedPV(
            devname + ':DiagStatus-Mon', _SIRFStatusPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        self._pvs_connected = {pv: False for pv in self.pvs}

    def _update_pvs(self):
        for pvo in self.pvs:
            if not pvo.connected:
                if self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.TIMEOUT_ALARM,
                        severity=_Severity.INVALID_ALARM,
                        field='status')
                self._pvs_connected[pvo] = False
                if 'DiagStatus' in pvo.pvname:
                    self.run_callbacks(pvo.pvname, value=pvo.value)
            else:
                if not self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.NO_ALARM,
                        severity=_Severity.NO_ALARM, field='status')
                self._pvs_connected[pvo] = True
                self.run_callbacks(pvo.pvname, value=pvo.value)
